import numpy as np
import inspect
import sys


small = .000000000001


class NeighborhoodFactory(object):

    """
    Class for creating the neighborhood function definition object.
    """

    @staticmethod
    def build(neighborhood_func):
        for name, obj in inspect.getmembers(sys.modules[__name__]):
            if inspect.isclass(obj):
                if hasattr(obj, 'name') and neighborhood_func == obj.name:
                    return obj()
        else:
            raise Exception(
                "Unsupported neighborhood function '%s'" % neighborhood_func)


class GaussianNeighborhood(object):

    """
    Creation of a Gaussian neighborhood function definition object.
    """

    name = 'gaussian'

    @staticmethod
    def calculate(distance_matrix, radius, dim):
        return np.exp(-1.0*distance_matrix/(2.0*radius**2)).reshape(dim, dim)

    def __call__(self, *args, **kwargs):
        return self.calculate(*args)


class BubbleNeighborhood(object):

    name = 'bubble'

    @staticmethod
    def calculate(distance_matrix, radius, dim):
        def l(a, b):
            c = np.zeros(b.shape)
            c[a-b >= 0] = 1
            return c

        return l(radius,
                 np.sqrt(distance_matrix.flatten())).reshape(dim, dim) + small

    def __call__(self, *args, **kwargs):
        return self.calculate(*args)
    


class NormalizerFactory(object):
    """
    Class for creating the data normalization object.
    """

    @staticmethod
    def build(type_name):
        for name, obj in inspect.getmembers(sys.modules[__name__]):
            if inspect.isclass(obj):
                if hasattr(obj, 'name') and type_name == obj.name:
                    return obj()
        else:
            raise Exception("Unknown normalization type '%s'" % type_name)


class Normalizer(object):

    """
    Class model for implementing new normalization objects.
    """

    name = "Object_name"

    def normalize(self, data):
        raise NotImplementedError()

    def normalize_by(self, raw_data, data):
        raise NotImplementedError()

    def denormalize_by(self, raw_data, data):
        raise NotImplementedError()

# Para criar outros normalizadores só copiar esse codigo e substituir
class VarianceNormalizer(Normalizer):

    """
    Normalization by variance. The data will be normalized by subtracting the mean and dividing by 
    the standard deviation of that variable.
    """

    name = 'var'

    def _mean_and_standard_dev(self, data):
        return np.nanmean(data, axis=0), np.nanstd(data, axis=0)

    def normalize(self, data):
        me, st = self._mean_and_standard_dev(data)
        st[st == 0] = 1  # prevent: when sd = 0, normalized result = NaN
        return np.round((data-me)/st,10)

    def normalize_by(self, raw_data, data, with_labels=False, pred_size=None):
        if with_labels:
            me, st = self._mean_and_standard_dev(raw_data[:, : -pred_size])
            st[st == 0] = 1  # prevent: when sd = 0, normalized result = NaN

        else:
            me, st = self._mean_and_standard_dev(raw_data)
            st[st == 0] = 1  # prevent: when sd = 0, normalized result = NaN

        return np.round((data - me) / st,10)

    def denormalize_by(self, data_by, n_vect, with_labels=False, pred_size=None):
        if with_labels:
            me, st = self._mean_and_standard_dev(data_by[:, :(data_by.shape[1] - pred_size)])
        else:
            me, st = self._mean_and_standard_dev(data_by)
        return np.round(n_vect * st + me,10)
    
class NoneNormalizer(Normalizer):
    """
    Class for handling cases when normalization should not be applied.
    """

    name = "None"

    def normalize(self, data):
        return data

    def normalize_by(self, raw_data, data):
        return data

    def denormalize_by(self, raw_data, data):
        return data





from logging.config import dictConfig
import matplotlib

dictConfig({
    "version": 1,
    "disable_existing_loggers": True,
    "root": {
        "level": "NOTSET",
        "handlers": ["console"]
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "basic"
        }
    },
    "formatters": {
        "basic": {
            "format": '%(message)s'
        }
    }
})



from .intrasom import SOMFactory



import numpy as np
from sklearn.decomposition import PCA

class Codebook(object):
    """
    Class for creating the SOM codebook. The codebook is the matrix of weights that is
    trained in its competitive, collaborative and adaptive synaptic processes.
    """

    def __init__(self, mapsize, lattice, mapshape, dist_factor):
        self.lattice = lattice
        self.mapshape = mapshape
        self.dist_factor = dist_factor

        if 2 == len(mapsize):
            _size = [1, np.max(mapsize)] if 1 == np.min(mapsize) else mapsize

        elif 1 == len(mapsize):
            _size = [1, mapsize[0]]
            print('Sizemap input was considered\
                as the number of neural network nodes')
            print('The size of the map \
                is [{dlen},{dlen}]'.format(dlen=int(mapsize[0] / 2)))
        else:
            pass

        self.mapsize = _size
        self.nnodes = self.mapsize[0] * self.mapsize[1]
        self.matrix = np.asarray(self.mapsize)
        self.initialized = False


    @property
    def get_matrix(self):
        """
        Returns the array stored in the current object.

        Returns:

        matrix: ndarray
        The array stored in the current object.
        """
        return self.matrix

    def random_initialization(self, data):
        """
        Random type initialization
        Args:
            data: data used for initialization

        Returns:
            Array initialized with the same dimensions as the input data.
        """

        # Constructs an array by repeating the smallest and largest data values ​​with
        # extension being the number of nodes required
        # Modification for nanmin and nanmax to take into account the
        # existence of NaN data
        mn = np.tile(np.nanmin(data, axis=0), (self.nnodes, 1))
        mx = np.tile(np.nanmax(data, axis=0), (self.nnodes, 1))

        # Minimum value + (maximum value - minimum value)*(array of random
        # numbers [0,1] with the format of number of nodes of lines and number of
        # columns of data in columns. A training matrix randomly
        # initialized.
        #np.random.seed(0)
        self.matrix = mn + (mx - mn) *\
            (np.random.rand(self.nnodes, data.shape[1]))
        self.initialized = True

    def pca_linear_initialization(self, data):
        """
        Initialization of the map by using just the first two eigenvalues and
        eigenvectors. The initialization is done in the following steps:
        1. Transformation of input data and creation of a template matrix
        2. Creation of a matrix with scaling factors for each principal component (PC)
        3. Obtaining and normalization of the eigenvector(s)
        4. Linear combination for each node of the template matrix with the PC

        Args:
            data: data to use for the initialization

        Returns:
            Initialized matrix with same dimension as input data.
        """
        cols = self.mapsize[1]
        coord = None
        pca_components = None

        me = np.mean(data, 0)
        data = (data - me)
        tmp_matrix = np.tile(me, (self.nnodes, 1))

        if np.min(self.mapsize) > 1:
            coord = np.zeros((self.nnodes, 2))
            pca_components = 2

            for i in range(0, self.nnodes):
                coord[i, 0] = int(i / cols)  # x
                coord[i, 1] = int(i % cols)  # y

        elif np.min(self.mapsize) == 1:
            coord = np.zeros((self.nnodes, 1))
            pca_components = 1

            for i in range(0, self.nnodes):
                coord[i, 0] = int(i % cols)  # y

        mx = np.max(coord, axis=0)
        mn = np.min(coord, axis=0)
        coord = (coord - mn)/(mx-mn)
        coord = (coord - .5)*2

        pca = PCA(n_components=pca_components, svd_solver='randomized')

        pca.fit(data)
        eigvec = pca.components_
        eigval = pca.explained_variance_
        norms = np.sqrt(np.einsum('ij,ij->i', eigvec, eigvec))
        eigvec = ((eigvec.T/norms)*eigval).T

        for j in range(self.nnodes):
            for i in range(eigvec.shape[0]):
                tmp_matrix[j, :] = tmp_matrix[j, :] + coord[j, i]*eigvec[i, :]

        self.matrix = np.around(tmp_matrix, decimals=6)
        self.initialized = True
    
    def pretrain(self):
        self.initialized = True

    def pretrain(self):
        self.initialized = True


    def grid_dist(self, node_ind):
        """
        Calculates distances on the grid for maps with planar or toroidal
        topology and with rectangular or hexagonal lattice.

        Args:
            node_ind: neural network node index, between 0 and nnodes-1.

        Returns:
            Returns the distances from this node to all other grid nodes,
            within the parameters specified in the SOM object.

        """

        # Define which function to call for each lattice and topology
        if self.mapshape == 'planar':
            if self.lattice == 'rect':
                return self._rect_dist_plan(node_ind)

            elif self.lattice == 'hexa':
                return self._hexa_dist_plan(node_ind)

        if self.mapshape == 'toroid':
            if self.lattice == 'rect':
                return self._rect_dist_tor(node_ind)

            elif self.lattice == 'hexa':
                return self._hexa_dist_tor(node_ind)

    def _rect_dist_plan(self, node_ind):
        """
        Finds the Manhattan distance matrix (L1) of a neural network node
        for all others, for a rectangular lattice in a map with
        planar topology.

        Args:
            node_ind: neural network node index, between 0 and nnodes-1.

        Returns:
            Returns array of distances from this node to all other nodes in the
            grid, on a planar map.

        """
        # Separate column and row values
        rows, cols = self.mapsize

        # Generate the xy coordinates of the BMUs for a rectangular grid
        coordinates = self.generate_rec_lattice(rows, cols)

        # Find the Manhattan distances for a rectangular grid through
        # its coordinates
        dist = np.array(abs(coordinates[ind] - coordinates[node_ind]).sum() \
            for ind in range(len(coordinates)))

        return dist

    def _rect_dist_tor(self, node_ind):
        """
        Finds the matrix of distances from a neural network node to all
        others, for a hexagonal lattice in a map with toroidal topology.
        Args:
            node_ind: Neural network node index, between 0 and nnodes-1.

        Returns:
            Returns the distances from this node to all other grid nodes, in a
            toroidal map.

        """
        rows, cols = self.mapsize

        # Generate the xy coordinates of the BMUs for a rectangular grid
        coordinates = self.generate_hex_lattice(rows, cols)

        # Extends the distance search to the neighborhood created by the toroidal topology, creating periodicity of the data
        toroid_neigh = [[0, 0], [cols, 0], [cols, rows], [0, -rows],
            [-cols, 0], [0, -rows], [-cols, -rows]]

        # Calculate the distances in the toroidal topology, finding all possible distances according to toroid_neigh and
        # selecting the smallest
        dist = np.array(
            [min([abs((coordinates[ind] + [neig]) - coordinates[node_ind]).sum()\
             for neig in toroid_neigh]) for ind in range(len(coordinates))])

        return dist

    def _hexa_dist_plan(self, node_ind):
        """

        Finds the matrix of distances from a neural network node to all
        others, for a hexagonal lattice in a map with planar topology.

        Args:
            node_ind: neural network node index, between 0 and nnodes-1.

        Returns:
            Returns the distances from this node to all other grid nodes, in a
            planar map.

        """
        cols, rows = self.mapsize

        # Generate x,y coordinates for a hexagonal grid
        coordinates = self.generate_oddr_cube_lattice(cols, rows)

        # Find the manhatan distances for a hexagonal grid via their xy coordinates
        dist = np.array([self.cube_distance(coordinates[node_ind], coordinates[i], dist_factor=self.dist_factor)\
         for i in range(len(coordinates))])

        return dist.astype(int)

    def _hexa_dist_tor(self, node_ind):
        """
        Finds the Manhattan distance matrix (L1) of a neural network node
        for all others, for a rectangular lattice in a map with
        toroidal topology.

        Args:
            node_ind: Neural network node index, between 0 and nnodes-1.

        Returns:
            Returns an array of distances from this node to all other nodes in the
            grid, in a toroidal map.
        """

        # Separate column and row values
        cols, rows = self.mapsize

        # Generate the BMUs xyz coordinates for a hexagonal grid
        coordinates = self.generate_oddr_cube_lattice(cols, rows)
        
        
        # Extend the distance search to the neighborhood created by the toroidal
        # topology, creating data periodicity
        toroid_neigh = self.toroid_neighborhood(cols, rows)

        # Calculate the distances in the toroidal topology, finding all
        # possible distances according to toroid_neigh and selecting the smallest
        dist = np.zeros(((cols*rows),9))
        for i in range(cols*rows):
            if i >= node_ind:
                for j, neig in enumerate(toroid_neigh):
                    dist[i,j] = self.cube_distance(coordinates[i]+neig, coordinates[node_ind], dist_factor=self.dist_factor)


        return np.min(dist, axis=1).astype(int)

    def toroid_neighborhood(self, cols, rows):
        """
        Function to generate the cubic coordinates in the toroidal neighborhood for a given
        map size, in order: [Center, Right, Bottom Right, Bottom,
        Bottom Left, Left, Top Left, Top, Top Right].

        Args:

            cols: number of map columns to be generated.

            rows: number of rows of the map to be generated.

        Returns:
            List of distance coordinates
        """
        toroid_neigh = [[0, 0],
                        [cols, 0],
                        [cols, rows],
                        [0, rows],
                        [-cols, rows],
                        [-cols, 0],
                        [-cols, -rows],
                        [0,-rows],
                        [cols, -rows]]

        return [self.oddr_to_cube(i[0], i[1]) for i in toroid_neigh]


    def generate_oddr_cube_lattice(self, n_columns, n_rows):
        """
        Function to generate cubic coordinates in [x,y,z] format for an odd-r
        hexagonal grid (odd lines shifted to the right) for a
        predetermined number of columns and row.
        Args:
            n_columns: number of columns

            n_rows: number of rows

        Returns:
            coordinates: list[x, y, z]
        """
        x_coord = []
        y_coord = []
        z_coord = []
        for j in range(n_rows):
            for i in range(n_columns):
                x = i-(j-(j & 1))/2
                z = j
                y = -x -z

                # Put in lists
                x_coord.append(x)
                y_coord.append(y)
                z_coord.append(z)

        coordinates = np.column_stack([x_coord, y_coord, z_coord])
        return coordinates

    def cube_distance(self, a, b, dist_factor = 2):
        """
        Calculates the Euclidean distance between two cubic coordinates
        Args:
            a: First cubic coordinate [x,y,z]
            b: Second cubic coordinate [x,y,z]

        Returns:

            Manhattan distance between coordinates.
        """
        return ((abs(a[0] - b[0]) + abs(a[1] - b[1]) + abs(a[2] - b[2]))/2)**dist_factor

    def oddr_to_cube(self, col, row):
        """
        Transforms rectangular coordinates to cubic.

        Args:
            coord: Coordinate you want to transform.

        Returns:
            Cubic coordinate in [x,y,z] format
        """

        x = col - (row - (row & 1)) / 2
        z = row
        y = -x-z
        return [x, y, z]


    def generate_rec_lattice(self, n_rows, n_columns):
        """
        Generates the xy coordinates of the BMUs for a rectangular grid.

        Args:
            n_rows: Number of rows in the Kohonen map.
            n_columns: Number of columns in the Kohonen map.

        returns:
            Coordinates in the [x,y] format for the BMUs in a rectangular grid.
        """
        x_coord = []
        y_coord = []
        for j in range(n_rows):
            for i in range(n_columns):
                x_coord.append(i)
                y_coord.append(j)
        coordinates = np.column_stack([x_coord, y_coord])
        return coordinates

    def generate_hex_lattice(self, n_columns, n_rows):
        """
        Generates the xy coordinates of the BMUs for an odd-r hexagonal grid.
        Args:
            n_rows: Number of rows in the Kohonen map.
            n_columns: Number of columns in the Kohonen map.

        Returns:
            Coordinates in the [x,y] format for the bmus in a hexagonal grid.
        """
        ratio = np.sqrt(3) / 2

        coord_x, coord_y = np.meshgrid(np.arange(n_columns),
                                        np.arange(n_rows), 
                                        sparse=False, 
                                        indexing='xy')
        coord_y = coord_y * ratio
        coord_x = coord_x.astype(float)
        coord_x[1::2, :] += 0.5
        coord_x = coord_x.ravel()
        coord_y = coord_y.ravel()

        coordinates = np.column_stack([coord_x, coord_y])

        return coordinates



import numpy as np
from sklearn.preprocessing import minmax_scale
from numpy import nan, dot, nanmean
from math import sqrt
import matplotlib.pyplot as plt
from matplotlib.patches import RegularPolygon
from matplotlib import cm
import matplotlib.gridspec as gridspec
import matplotlib as mpl
from tqdm.auto import tqdm
from scipy import sparse as sp
from PIL import Image
import glob
import os
from textwrap import fill
from numpy import pi, sin, cos
from sklearn.cluster import KMeans
from sklearn.utils import shuffle
import pkg_resources
import plotly.graph_objs as go
from scipy.ndimage import rotate
from skimage.transform import resize
from sklearn.metrics.pairwise import nan_euclidean_distances
import statistics



class PlotFactory(object):

    def __init__(self, som_object):
        self.name = som_object.name
        self.codebook = som_object.codebook.matrix
        self.mapsize = som_object.mapsize
        self.bmus = som_object._bmu[0].astype(int)
        self.neuron_matrix = som_object.neuron_matrix
        self.component_names = som_object._component_names
        self.sample_names = som_object._sample_names
        self.unit_names = som_object._unit_names
        self.rep_sample = som_object.rep_sample
        self.data_denorm = som_object.denorm_data(som_object._data)
        self.data_proj_norm = som_object.data_proj_norm
        ## added
        self.build_umatrix = som_object.build_umatrix
        
        # Load foot image
        image_file = pkg_resources.resource_filename('intrasom', 'images/foot.jpg')
        self.foot = Image.open(image_file)

    # def build_umatrix(self, expanded=False, log=False):
    #     """
    #     Function to calculate the U-Matrix of unified distances from the trained weight matrix.
    # 
    #     Args:
    #         exapanded: Boolean value to indicate whether the return will be the summarized 
    #             U-Matrix (average distances of the 6 neighboring BMUs) or the expanded 
    #             U-Matrix (all distance values).
    #         
    #         log: Returns the base 10 logarithm for the distance values. It is used when 
    #             there are samples with a large dissimilarity boundary that masks the 
    #             visualization of the U-Matrix. The logarithmic transformation of these 
    #             values allows for a better visualization of the matrix.
    # 
    #     Returns:
    #         Expanded or summarized U-Matrix of distances.
    #     """
    #     # Function to find distance quickly
    #     def fast_norm(x):
    #         """
    #         Returns the L2 norm of a 1-D array.
    #         """
    #         return sqrt(dot(x, x.T))
    # 
    #     print('Using visualization.py definition')
    #     # Neurons weights matrix
    #     weights = np.reshape(self.codebook, (self.mapsize[1], self.mapsize[0], self.codebook.shape[1]))
    # 
    #     # Neighbor hexagonal search
    #     ii = [[1, 1, 0, -1, 0, 1], [1, 0,-1, -1, -1, 0]]
    #     jj = [[0, 1, 1, 0, -1, -1], [0, 1, 1, 0, -1, -1]]
    # 
    #     # Initialize U Matrix
    #     um = np.nan * np.zeros((weights.shape[0], weights.shape[1], 6))
    # 
    #     # Fill U-matrix
    #     for y in range(weights.shape[0]):
    #         for x in range(weights.shape[1]):
    #             w_2 = weights[y, x]
    #             e = y % 2 == 0
    #             for k, (i, j) in enumerate(zip(ii[e], jj[e])):
    #                 if (x+i >= 0 and x+i < weights.shape[1] and y+j >= 0 and y+j < weights.shape[0]):
    #                     w_1 = weights[y+j, x+i]
    #                     um[y, x, k] = fast_norm(w_2-w_1)
    #     if expanded:
    #         # Expanded U Matrix
    #         return np.log(um) if log else um
    #     else:
    #         # Reduced U Matrix
    #         return nanmean(np.log(um), axis=2) if log else nanmean(um, axis=2)
                        
    def plot_umatrix(self,
                     figsize = (10,10),
                     hits = True,
                     title = "U-Matrix",
                     title_size = 40,
                     title_pad = 25,
                     legend_title = "Distance",
                     legend_title_size = 25,
                     legend_ticks_size = 20,
                     save = True,
                     watermark_neurons = False,
                     watermark_neurons_alfa = 0.5,
                     neurons_fontsize = 7,
                     file_name = None,
                     file_path = False, 
                     resume = False,
                     label_plot = False, 
                     label_plot_name = None,
                     project_samples_label = None,
                     samples_label = False,
                     samples_label_index = None,
                     samples_label_fontsize = 8,
                     save_labels_rep = False,
                     label_title_xy = (0,0.5),
                     log=False):
        """
        Function to plot the U-Matrix of unified distances.

        Args:
            figsize:  Size of the U-Matrix plotting area. The adjustment of these values is 
                crucial for the proper distribution of the plotted objects and strongly depends 
                on the shape of the trained map (number of rows and columns). Default: (10, 10)

            hits: Boolean value to indicate whether to plot the hits of input vectors on the 
                BMUs (proportional to the number of vectors per BMU). These hits are visualized as 
                white hexagons with size proportional to the number of input samples represented 
                by that BMU.

            title: Title of the created figure. Default: "U-Matrix"

            title_size: Size of the plotted title. Default: 40

            title_pad: Spacing between the title and the top of the matrix. Default: 25

            legend_title: Title of the legend color bar. Default: "Distance"

            legend_title_size: Size of the legend title. Default: 25

            legend_ticks_size: Size of the digits on the legend color bar. Default: 20

            save: Boolean value to define whether to save the created image. The image will be saved in the 
                directory (Plotagens/Matriz_U). Default: True

            watermark_neurons: Boolean value to add a watermark with the neuron numbers to the image. 
                Default: False

            watermark_neurons_alfa: Value between 0 and 1 indicating the transparency of the plotted 
                neuron template over the U-Matrix. The closer to 1, the lower the transparency. Default: 0.5.

            neurons_fontsize: Font size for the neuron numbers to be plotted. Default: 7.

            file_name: Name to be given to the saved file. If no name is provided, the project name 
                will be used.

            file_path: System path where the image should be saved, if a custom path is preferred.

            resume: Boolean value to plot only the upper part of the U-Matrix, omitting the mirrored 
                lower part. Default: False

            label_plot: Boolean value to add labels to the hexagons of the U-Matrix according to a 
                boolean variable present in the training. Default: False

            label_plot_names: Name of the variable for plotting. The presence of the variable is 
                indicated by white hits, and the absence by black hits. Default: None

            samples_label: Boolean value to indicate the plotting of labels for selected samples. 
                Default: False.

            samples_label_index: List of indices of the selected samples for label plotting. 
                Default: None. "All" for plotting all samples.

            samples_label_fontsize: Font size for the plotted labels. Default: 8.

            save_labels_rep: Boolean value to indicate whether to save a .txt file in Results with 
                the selected samples and their representativeness order in each BMU. Default: False.

            label_title_xy: Coordinates (x, y) to position the label title. Default: (-0.02, 1.1)

            log: Boolean value to plot the U-Matrix on a logarithmic scale for better visualization 
                of dissimilarity boundaries in the presence of outliers.

        Returns:
            The image with the plot of the U-Matrix of unified distances.
        """
        def select_keys(original_dict, keys_list):
            new_dict = {}
            keys_list = np.array(keys_list)
            if keys_list.shape[0]>1:
                for key in keys_list:
                    if key in original_dict:
                        new_dict[key] = original_dict[key]
            else:
                if keys_list[0] in original_dict:
                    new_dict[keys_list[0]] = original_dict[keys_list[0]]
            return new_dict
        
        def search_strings(search_list, target_list):
            found_index = None
            found_string = None

            for search_string in search_list:
                for index, target_string in enumerate(target_list):
                    if search_string == target_string:
                        if found_index is None or index < found_index:
                            found_index = index
                            found_string = search_string
                        break

            return found_string, found_index

        if file_name is None:
            file_name = f"U_Matrix_{self.name}"

        if hits:
            bmu_dic = self.hits_dictionary

        # Create coordinates
        xx = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,0], (self.mapsize[1], self.mapsize[0]))
        yy = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,1], (self.mapsize[1], self.mapsize[0]))

        # U Matrix
        um = self.build_umatrix(expanded = True, log=log)
        umat = self.build_umatrix(expanded = False, log=log)
        
        
        if resume:
            # Plotting
            prop = self.mapsize[1]*0.8660254/self.mapsize[0]
            f = plt.figure(figsize=(5, 5*prop), dpi=300)
            f.patch.set_facecolor('blue')
            ax = f.add_subplot()
            ax.set_aspect('equal')

            # Normalize colors for all hexagons
            norm = mpl.colors.Normalize(vmin=np.nanmin(um), vmax=np.nanmax(um))
            counter = 0

            for j in range(self.mapsize[1]):
                for i in range(self.mapsize[0]):
                    # Central Hexagon
                    hex = RegularPolygon((xx[(j, i)]*2,
                                          yy[(j,i)]*2),
                                         numVertices=6,
                                         radius=1/np.sqrt(3),
                                         facecolor= cm.jet(norm(umat[j][i])),
                                         alpha=1)#, edgecolor='black')

                    ax.add_patch(hex)

                    # Upper Right Hexagon
                    if not np.isnan(um[j, i, 0]):
                        hex = RegularPolygon((xx[(j, i)]*2+1,
                                              yy[(j,i)]*2),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,0])),
                                             alpha=1)
                        ax.add_patch(hex)

                    # Upper Left Hexagon
                    if not np.isnan(um[j, i, 1]):
                        hex = RegularPolygon((xx[(j, i)]*2+0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,1])),
                                             alpha=1)
                        ax.add_patch(hex)

                    # Hexagono Superior Esquerdo
                    if not np.isnan(um[j, i, 2]):
                        hex = RegularPolygon((xx[(j, i)]*2-0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,2])),
                                             alpha=1)
                        ax.add_patch(hex)
                        

                    
                    if j==0:
                        # Central Hexagon
                        hex = RegularPolygon((xx[(j, i)]*2,
                                              yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor= cm.jet(norm(umat[j][i])),
                                             alpha=1)#, edgecolor='black')
                        ax.add_patch(hex)
                        
                        # Right
                        if not np.isnan(um[j, i, 0]):
                            hex = RegularPolygon((xx[(j, i)]*2+1,
                                                  yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,0])),
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                        # Lower Right Hexagon
                        if not np.isnan(um[j, i, 1]):
                            hex = RegularPolygon((xx[(j, i)]*2+0.5,
                                                  yy[(j,i)]*2-(np.sqrt(3)/2)+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,1])),
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                        # Bottom Left Hexagon
                        if not np.isnan(um[j, i, 2]):
                            hex = RegularPolygon((xx[(j, i)]*2-0.5,
                                                  yy[(j,i)]*2-(np.sqrt(3)/2)+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,2])),
                                                 alpha=1)
                            ax.add_patch(hex)
                    if i==0:
                        # Central Hexagon
                        hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2 - 1,
                                              yy[(j,i)]*2),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor= cm.jet(norm(umat[j][i])),
                                             alpha=1)#, edgecolor='red')
                        ax.add_patch(hex)
                        
                        # Right
                        if not np.isnan(um[j, i, 0]):
                            hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2,
                                                  yy[(j,i)]*2),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,0])),
                                                 alpha=1)
                            ax.add_patch(hex)
                                                 
                        # Upper Right Hexagon
                        if not np.isnan(um[j, i, 1]):
                            hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2 - 0.5,
                                                  yy[(j,i)]*2+(np.sqrt(3)/2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,1])),
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                        # Bottom Left Hexagon
                        if not np.isnan(um[j, i, 2]):
                            hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2 - 1.5,
                                                  yy[(j,i)]*2+(np.sqrt(3)/2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,2])),
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                    if i==0 and j==0:
                        # Central Hexagon
                        hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2-1,
                                              yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor= cm.jet(norm(umat[j][i])),
                                             alpha=1)#, edgecolor='black')
                        ax.add_patch(hex)
                        
                        # Right
                        if not np.isnan(um[j, i, 0]):
                            hex = RegularPolygon((xx[(j, i)]*2 + self.mapsize[0]*2,
                                                  yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,0])),
                                                 alpha=1)
                            ax.add_patch(hex)
                                                 
                       # Lower Right Hexagon
                        if not np.isnan(um[j, i, 1]):
                            hex = RegularPolygon((xx[(j, i)]*2+ self.mapsize[0]*2-0.5,
                                                  yy[(j,i)]*2-(np.sqrt(3)/2)+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,1])),
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                        # Bottom Left Hexagon
                        if not np.isnan(um[j, i, 2]):
                            hex = RegularPolygon((xx[(j, i)]*2+ self.mapsize[0]*2-1.5,
                                                  yy[(j,i)]*2-(np.sqrt(3)/2)+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3),
                                                 facecolor=cm.jet(norm(um[j,i,2])),
                                                 alpha=1)
                            ax.add_patch(hex)

                    # Plot hits
                    if hits:
                        try:
                            hex = RegularPolygon((xx[(j,i)]*2,
                                                  yy[(j,i)]*2),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3)*bmu_dic[counter],
                                                 facecolor='white',
                                                 edgecolor='lightgray',
                                                 linewidth=1,
                                                 alpha=1)
                            ax.add_patch(hex)
                            
                            if j==0:
                                hex = RegularPolygon((xx[(j,i)]*2,
                                                  yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3)*bmu_dic[counter],
                                                 facecolor='white',
                                                 edgecolor='lightgray',
                                                 linewidth=1,
                                                 alpha=1)
                                ax.add_patch(hex)
                                
                            if i==0:
                                hex = RegularPolygon((xx[(j,i)]*2 + self.mapsize[0]*2 - 1,
                                                  yy[(j,i)]*2),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3)*bmu_dic[counter],
                                                 facecolor='white',
                                                 edgecolor='lightgray',
                                                 linewidth=1,
                                                 alpha=1)
                                ax.add_patch(hex)
                            
                            if i==0 and j==0:
                                hex = RegularPolygon((xx[(j,i)]*2 + self.mapsize[0]*2 - 1,
                                                  yy[(j,i)]*2+(0.8660254*self.mapsize[1]*2)),
                                                 numVertices=6,
                                                 radius=1/np.sqrt(3)*bmu_dic[counter],
                                                 facecolor='white',
                                                 edgecolor='lightgray',
                                                 linewidth=1,
                                                 alpha=1)
                                ax.add_patch(hex)
                        except:
                            pass
                            
                    counter+=1

            # Plot Parameters
            
            plt.xlim(0.5, 2*self.mapsize[0]-0.5)
            plt.ylim(0, 2*self.mapsize[1]*0.8660254)

            f.tight_layout()
            ax.set_axis_off()
            plt.gca().invert_yaxis()
            plt.close()

            os.makedirs("Plots/U_matrix", exist_ok=True)


            filename = "Plots/U_matrix/plain_umat.jpg"
            f.savefig(filename, dpi=300, bbox_inches = "tight")

            # Read the saved image into a NumPy ndarray
            umat_plain = mpl.image.imread(filename)[30:-30,30:-30]
            
            
            return umat_plain
 
        else:
            # Plotting
            f = plt.figure(figsize=figsize, dpi=300)
            
            gs = gridspec.GridSpec(100, 100)
            ax = f.add_subplot(gs[:95, 0:90])
            ax.set_aspect('equal')

            # Normalize colors for all hexagons
            norm = mpl.colors.Normalize(vmin=np.nanmin(um), vmax=np.nanmax(um))
            counter = 0
            
            if watermark_neurons:
                # For use to plot number of BMUs
                nnodes = self.mapsize[0] * self.mapsize[1]
                grid_bmus = np.linspace(1,nnodes, nnodes).reshape(self.mapsize[1], self.mapsize[0])

            for j in range(self.mapsize[1]):
                for i in range(self.mapsize[0]):
                    # Central Hexagon
                    hex = RegularPolygon((xx[(j, i)]*2,
                                          yy[(j,i)]*2),
                                         numVertices=6,
                                         radius=1/np.sqrt(3),
                                         facecolor= cm.jet(norm(umat[j][i])),
                                         alpha=1)#, edgecolor='black')

                    ax.add_patch(hex)

                     # Right Hexagon
                    if not np.isnan(um[j, i, 0]):
                        hex = RegularPolygon((xx[(j, i)]*2+1,
                                              yy[(j,i)]*2),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,0])),
                                             alpha=1)
                        ax.add_patch(hex)

                    # Upper Right Hexagon
                    if not np.isnan(um[j, i, 1]):
                        hex = RegularPolygon((xx[(j, i)]*2+0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,1])),
                                             alpha=1)
                        ax.add_patch(hex)

                    # Upper Left Hexagon
                    if not np.isnan(um[j, i, 2]):
                        hex = RegularPolygon((xx[(j, i)]*2-0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,2])),
                                             alpha=1)
                        ax.add_patch(hex)

                    if watermark_neurons:
                        # Central Hexagon
                        hex = RegularPolygon((xx[(j, i)]*2,
                                              yy[(j,i)]*2),
                                              numVertices=6,
                                              radius=2/np.sqrt(3),
                                              facecolor= "white",
                                              edgecolor='black',
                                              alpha=watermark_neurons_alfa)#, edgecolor='black')
                        ax.add_patch(hex)

                        ax.text(xx[(j,i)]*2, yy[(j,i)]*2, 
                                s=f"{int(grid_bmus[j,i])}", 
                                size = neurons_fontsize,
                                horizontalalignment='center', 
                                verticalalignment='center', 
                                color='black')

                    # Plot hits
                    if hits:
                        if label_plot:
                            ind_var = list(self.component_names).index(label_plot_name)
                            bool_val = self.data_denorm[:,ind_var]
                            bool_val = np.array([0 if number < 0.5 else 1 for number in bool_val])

                            if counter in self.bmus:
                                vars_in_hit =  np.array([ind for ind, value in enumerate(self.bmus) if value == counter])
                                bool_hit = bool_val[vars_in_hit]
                                bool_hit = statistics.mode(bool_hit)

                                facecolor_hits = "black" if bool_hit == 0 else 'white'
                            else:
                                pass
                        else:
                            facecolor_hits = 'white'

                        try:
                            hex = RegularPolygon((xx[(j, i)]*2,
                                                  yy[(j,i)]*2),
                                                 numVertices=6,
                                                 radius=((1/np.sqrt(3))*bmu_dic[counter]),
                                                 facecolor=facecolor_hits,
                                                 edgecolor=facecolor_hits,
                                                 linewidth=1.1,
                                                 alpha=1)
                            ax.add_patch(hex)
                        except:
                            pass
                    counter+=1

            if samples_label:
                if project_samples_label is not None:
                    samples_label_names = project_samples_label.index.tolist()
                    som_bmus = (project_samples_label.BMU.values-1).tolist()
                    rep_samples_dic = select_keys(self.rep_sample(project = project_samples_label), som_bmus)
                else:
                    if samples_label_index == "all":
                        samples_label_names = self.sample_names
                        samples_label_index = np.arange(0, len(samples_label_names), 1)
                    else:
                        samples_label_index = np.array(samples_label_index)
                        samples_label_names = self.sample_names[samples_label_index]
                    

                    som_bmus = self.bmus.astype(int)[samples_label_index]
                    rep_samples_dic = select_keys(self.rep_sample(), som_bmus+1)

                if save_labels_rep:
                    def filter_dictionary(dictionary, values):
                        filtered_dict = {}
                        for key, val in dictionary.items():
                            if isinstance(val, list):
                                matching_values = [item for item in val if item in values]
                                if matching_values:
                                    filtered_dict[key] = matching_values
                            else:
                                if val in values:
                                    filtered_dict[key] = val
                        return filtered_dict
                    
                    rep_samples_dic_filter = filter_dictionary(rep_samples_dic, samples_label_names)
                    with open('Results/Representative_samples_umatrix.txt', 'w', encoding='utf-8') as file:
                        for key, value in rep_samples_dic_filter.items():
                            if isinstance(value, list):
                                value = ', '.join(value)
                            file.write(f'BMU {key+1}: {value}\n')
                
                counter = 0
                for j in range(self.mapsize[1]):
                    for i in range(self.mapsize[0]):
                        try:
                            if isinstance(rep_samples_dic[counter+1], list):
                                total = len(rep_samples_dic[counter+1])
                                selected_samples = list(set(samples_label_names).intersection(set(rep_samples_dic[counter+1])))
                                sample_name, idx_sample = search_strings(selected_samples, rep_samples_dic[counter+1])
                                rep_sample_name = f"{sample_name}({idx_sample+1}/{total})"
                                                                
                            else:
                                rep_sample_name = rep_samples_dic[counter+1]
                            
                            hex = RegularPolygon((xx[(j, i)]*2,
                                                yy[(j,i)]*2),
                                                numVertices=6,
                                                radius=((1/np.sqrt(3))*0.5),
                                                facecolor='black',
                                                edgecolor='white',
                                                linewidth=1.1,
                                                alpha=1)
                            ax.add_patch(hex)

                            box_props = {'facecolor': 'white', 'edgecolor': 'black', 'boxstyle': 'round'}
                            ax.text(xx[(j,i)]*2, yy[(j,i)]*2+1.5*(1/np.sqrt(3)), 
                                    s=f"{rep_sample_name}", 
                                    size = samples_label_fontsize,
                                    horizontalalignment='center', 
                                    verticalalignment='top', 
                                    color='black',
                                    bbox= box_props)

                        except:
                            pass
                    
                        counter+=1
                    

            # Plot Parameters
            plt.xlim(-1.1, 2*self.mapsize[0]+0.1)
            plt.ylim(-0.5660254-0.6, 2*self.mapsize[1]*0.8660254-2*0.560254+0.6)
            ax.set_axis_off()
            plt.gca().invert_yaxis()

            # Map title
            plt.title(fill(title,30),
                      horizontalalignment='center',
                      verticalalignment='top',
                      size=title_size,
                      pad=title_pad)

            # Legend
            ax2 = f.add_subplot(gs[30:70, 95:98])
            cmap = mpl.cm.turbo
            norm = mpl.colors.Normalize(vmin=np.nanmin(um),
                                        vmax=np.nanmax(um))

            # Color bar
            cb1 = mpl.colorbar.ColorbarBase(ax2,
                                            cmap=cmap,
                                            norm=norm,
                                            orientation='vertical')
            # Legend Parameters
            cb1.ax.tick_params(labelsize=legend_ticks_size)

            
            cb1.set_label(fill(legend_title, 20), 
                          size=legend_title_size, 
                          labelpad=20)
            cb1.ax.yaxis.label.set_position(label_title_xy)
            """
            # Move the colorbar a little to the right
            pos = cb1.ax.get_position()
            pos.x0 += 0.08 * (pos.x1 - pos.x0)
            pos.x1 += 0.08 * (pos.x1 - pos.x0)
            cb1.ax.set_position(pos)
            
            cb1.ax.yaxis.label.set_position(label_title_xy)"""
            
            #ADD WATERMARK
            # Add white space subplot below the plot
            ax3 = f.add_subplot(gs[95:100, 0:20], zorder=-1)

            # Add the watermark image to the white space subplot
            ax3.imshow(self.foot, aspect='equal', alpha=1)
            ax3.axis('off')
            
            plt.subplots_adjust(bottom=3, top = 5,wspace=0)
            
            plt.show()
        
        if save:
            if file_path:
                f.savefig(f"{file_path}/{file_name}.jpg",dpi=300, bbox_inches = "tight")
            else:
                # Create directories if they don't exist
                path = 'Plots/U_matrix'
                os.makedirs(path, exist_ok=True)

                print("Saving.")
                if hits:
                    if label_plot:
                        f.savefig(f"Plots/U_matrix/{file_name}_with_hits_label.jpg",dpi=300, bbox_inches = "tight") 
                    elif watermark_neurons:
                        f.savefig(f"Plots/U_matrix/{file_name}_with_hits_watermarkneurons.jpg",dpi=300, bbox_inches = "tight") 
                    else:
                        f.savefig(f"Plots/U_matrix/{file_name}_with_hits.jpg",dpi=300, bbox_inches = "tight")
                else:
                    if label_plot:
                        f.savefig(f"Plots/U_matrix/{file_name}_with_label.jpg",dpi=300, bbox_inches = "tight") 
                    elif watermark_neurons:
                        f.savefig(f"Plots/U_matrix/{file_name}_watermarkneurons.jpg",dpi=300, bbox_inches = "tight") 
                    else:
                        f.savefig(f"Plots/U_matrix/{file_name}.jpg",dpi=300, bbox_inches = "tight")


    def component_plot(self,
                       component_name = 0,
                       figsize = (10,10),
                       title = None,
                       full_title = False,
                       title_size = 30,
                       title_pad = 25,
                       legend_title = False,
                       legend_pad = 0,
                       label_title_xy = (0,0.5),
                       legend_title_size = 24,
                       legend_ticks_size = 20,
                       save = False,
                       file_name = None,
                       file_path = False,
                       collage = False):
        """
        Function to perform the plotting of a trained variable map.

            Args:
                component_name: the name of the variable you want to plot. It is
                    accept the variable name in the form of a string or an integer
                    number related to the index of that variable in the initial
                    variables list.

                figsize: size of the variable's plot screen.
                    Default: (10,10)

                title: title of the created figure. Default: "U-Matrix"

                title_size: size of the plotted title. Default: 40

                title_pad: spacing between the title and the top of the map.
                    Default: 25

                legend_title: title of the color bar legend.
                    Default: "Distance"

                legend_pad: spacing between the legend title and the map.
                    Default: 0

                y_legend: spacing between legend title and bottom
                    of the figure. Default: 1.12

                legend_title_size = subtitle title size. Default: 24.

                legend_ticks_size = size of the color bar digits of the
                    subtitle. Default: 20.

                save: boolean to define the saving of the created image.
                    This saving will be done in the directory (Plots/Component_plots).
                    Default: True.

                file_name: the name that will be given to the saved file. If no name
                    is given, the name of the project will be used.

                file_path: the path on the system where the image should be
                    saved in case you do not choose to use the default path.

            Returns:
                The image plotting the variable map.
        """
        def check_path(filename):
            # Check if the file already exists in the directory
            if os.path.isfile(filename):
                # Extract the filename and extension
                name, ext = os.path.splitext(filename)

                # Append a number to the filename to make it unique
                i = 1
                while os.path.isfile(f"{name}_{i}{ext}"):
                    i += 1

                # Change the filename to the unique name
                filename = f"{name}_{i}{ext}"
            
            return filename
        
        if file_name is None:
            file_name = f"Component_plot_{self.name}"

        if isinstance(component_name, int):
            # Pegar uma variavel
            bmu_var = self.neuron_matrix[:,component_name].reshape(self.mapsize[1], self.mapsize[0])
            var_name = self.component_names[component_name]

            # Captar unidade da legenda
            if not legend_title:
                legend_title = self.unit_names[component_name]
                
        elif isinstance(component_name, str):
            index = list(self.component_names).index(component_name)
            bmu_var = self.neuron_matrix[:, index].reshape(self.mapsize[1], self.mapsize[0])
            var_name = self.component_names[index]
            # Captar unidade da legenda
            if not legend_title:
                legend_title = self.unit_names[index]
        else:
            print("Wrong name for component, accepts only string with component name or int with its position")

        # Create coordinates
        xx = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,0], (self.mapsize[1], self.mapsize[0]))
        yy = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,1], (self.mapsize[1], self.mapsize[0]))

        # Plotting
        f = plt.figure(figsize=figsize,dpi=300)
        gs = gridspec.GridSpec(100, 20)
        ax = f.add_subplot(gs[:95, 0:19])
        ax.set_aspect('equal')

        # Normalize colors for all hexagons
        norm = mpl.colors.Normalize(vmin=np.nanmin(bmu_var), vmax=np.nanmax(bmu_var))

        # Fill the plot with the hexagons
        for j in range(self.mapsize[1]):
            for i in range(self.mapsize[0]):
                ax.add_patch(RegularPolygon((xx[(j, i)], yy[(j,i)]),
                                     numVertices=6,
                                     radius=1/np.sqrt(3),
                                     facecolor=cm.jet(norm(bmu_var[j,i])),
                                     alpha=1))

        plt.xlim(-0.5, self.mapsize[0]+0.5)
        plt.ylim(-0.5660254, self.mapsize[1]*0.8660254+2*0.560254)
        ax.set_axis_off()
        plt.gca().invert_yaxis()

        # Map title
        if full_title:
            name = title if title is not None else var_name
        else:
            if title is not None:
                name = title
            else:
                name = var_name.split()
                name = f"{name[0]} {name[1]}" if len(name)>1 else f"{name[0]}"
        plt.title(fill(f"{name}",20), horizontalalignment='center',  verticalalignment='top', size=title_size, pad=title_pad)

        # Legend
        ax2 = f.add_subplot(gs[27:70, 19])
        cmap = mpl.cm.turbo
        norm = mpl.colors.Normalize(vmin=np.nanmin(bmu_var), vmax=np.nanmax(bmu_var))
        cb1 = mpl.colorbar.ColorbarBase(ax2,
                                        cmap=cmap,
                                        norm=norm,
                                        orientation='vertical')

        cb1.ax.tick_params(labelsize=legend_ticks_size)
        # Put the caption title if it has a variable name
        cb1.set_label(fill(legend_title,15),
                      size=legend_title_size,
                      labelpad=legend_pad,
                      horizontalalignment='right',
                      wrap=True)
        
        cb1.ax.yaxis.label.set_position(label_title_xy)
        if collage == False:
            #ADD WATERMARK
            # Add white space subplot below the plot
            ax3 = f.add_subplot(gs[95:100, 0:4], zorder=-1)

            # Add the watermark image to the white space subplot
            ax3.imshow(self.foot, aspect='equal', alpha=1)
            ax3.axis('off')
        
        plt.subplots_adjust(bottom=3, top = 5,wspace=0)
        if full_title:
            if isinstance(var_name, str):
                label_name = var_name.replace("/", "")
            else:
                label_name = var_name
        else:
            if isinstance(var_name, str):
                label_name = var_name[:7].replace(" ", "").replace(":", "")
            else:
                label_name = var_name[:6]

        if collage:
            path = 'Plots/Component_plots/Collage/temp'
            os.makedirs(path, exist_ok=True)
            path_name = check_path(f"Plots/Component_plots/Collage/temp/{label_name}.jpg")
            f.savefig(path_name,dpi=300, bbox_inches = "tight")

        if save:
            label_name = file_name if file_name is not None else label_name
            if file_path:
                path_name = check_path(f"{file_path}/{label_name}.jpg")
                f.savefig(path_name,dpi=300, bbox_inches = "tight")
            else:
                path = 'Plots/Component_plots'
                os.makedirs(path, exist_ok=True)
                path_name = check_path(f"Plots/Component_plots/{label_name}.jpg")
                f.savefig(path_name,dpi=300, bbox_inches = "tight")

    def multiple_component_plots(self,
                                wich = "all",
                                figsize = (10,10),
                                full_title = False,
                                title_size = 30,
                                title_pad = 25,
                                legend_title = "Presence",
                                legend_pad = 0,
                                label_title_xy = (0,0.5),
                                legend_title_size = 24,
                                legend_ticks_size = 20,
                                save = True,
                                file_path = False, 
                                collage = False):
        """
        Function for plotting a list or all variables trained in the
        SOM object.

        Args:
            wich: list of variables to be plotted and saved or "all" for
            plotting all variables.
        """
        if isinstance(wich, str):
            iterator = self.component_names
        else:
            iterator = wich

        # Iteração sobre a função de plotagem individual
        pbar = tqdm(iterator, mininterval=1)
        for name in pbar:
            pbar.set_description(f"Component: {name}")
            self.component_plot(component_name = name,
                               figsize = figsize,
                               full_title = full_title,
                               title_size = title_size,
                               title_pad = title_pad,
                               legend_title = legend_title,
                               legend_title_size = legend_title_size,
                               legend_ticks_size = legend_ticks_size,
                               legend_pad = legend_pad,
                               label_title_xy= label_title_xy,
                               save = save,
                               file_path = file_path,
                               collage = collage)
            plt.close()

        print("Finished")



    def component_plot_collage(self,
                               page_size = (2480, 3508), # A4 em pixels
                               grid = (4,4),
                               wich = "all",
                               figsize = (10,10),
                               full_title = False,
                               title_size = 30,
                               title_pad = 25,
                               legend_title = "Presence",
                               legend_title_size = 24,
                               legend_ticks_size = 20,
                               legend_pad = 0,
                               label_title_xy = (0,0.5),
                               file_path=False):
        """
        Function to create a collage of training component maps
        (Component Plots).

        Args:
            page_size: size of the plot collage page in pixels.
                Default: A4 (2480, 3508)

            grid: the format of the collage grid. Default: (4,4)

            wich: list of variables to be pasted and saved or "all" for
                plotting all variables.

            title_size: size of the plotted title. Default: 30

            title_pad: spacing between the title and the top of the array.
                Default: 25

            legend_title: title of the legend color bar.
                Default: "Presence"

            legend_pad: spacing between the legend title and the U matrix.
                Default: 0

            y_legend: spacing between legend title and bottom
                of the figure. Default: 1.12

            legend_title_size = subtitle title size. Default: 24.

            legend_ticks_size = size of the color bar digits of the
                subtitle. Default: 20.

            file_path = path to disired directory to save your collage.
        """
        # Support function only for using inside the function
        def resize_image(image, page, grid):
            """
           Function to resize the image on the sheet according to the defined
            grid.

            Args:
                image: the image that should be scaled in the grid.

                page: The page dimensions for the collage.

                grid: the format of the collage grid.
            """
            max_hor = page.size[0] / grid[1]
            max_ver = page.size[1] / grid[0]

            image.thumbnail((max_hor, max_ver))

            return image

        print("Generating maps...")
        # List component plots that haven't been made yet
        if isinstance(wich, str):
            if wich == "all":
                list_figs = "all"  
                n_components = len(self.component_names)
            else:
                print("The accepted parameters are 'all' or a list of variables or variable indexes.")
        if isinstance(wich, list):
            list_figs = wich
            n_components = len(wich)

        # Create component plots that have not been previously generated
        self.multiple_component_plots(wich = list_figs,
                                      figsize = figsize,
                                      full_title = full_title,
                                      title_size = title_size,
                                      title_pad = title_pad,
                                      save=False,
                                      legend_title = legend_title,
                                      legend_title_size = legend_title_size,
                                      legend_ticks_size = legend_ticks_size,
                                      legend_pad = legend_pad,
                                      label_title_xy = label_title_xy,
                                      collage = True)
        # Acquire the path for all images
        images_path = glob.glob('Plots/Component_plots/Collage/temp/*.jpg')

        # Number of pages needed to populate the components in the grid
        n_pages = int(n_components/(grid[0]*grid[1]))+1

        # Number of images per page
        im_pp = grid[0]*grid[1]

        print("Generating collage...")
        for i in range(n_pages):
            # Image to make the collage
            page = Image.new("RGB", page_size, "WHITE")  # White background

            # Slicing image paths for each page
            images_path_page = images_path[i*im_pp:(i+1)*im_pp]
            xx = np.tile(np.arange(grid[1]), grid[0])
            yy = np.repeat(np.arange(grid[0]), grid[1])

            for j, img_path in enumerate(images_path_page):
                img = Image.open(img_path) 
                img = resize_image(img, page, grid)
                x_pos = int(xx[j] * page_size[0] / grid[1])
                y_pos = int(yy[j] * page_size[1] / grid[0])
                page.paste(img, (x_pos, y_pos))

            # Save at plots root directory
            path = 'Plots/Component_plots/Collage/pages'
            os.makedirs(path, exist_ok=True)
            
            foot = self.foot
            max_height = page_size[1]/40
            width = int((foot.width / foot.height) * max_height)
            foot.thumbnail((width, max_height))
            x = 0  # Left position
            y = page.height - foot.height  # Down position
            page.paste(foot, (x, y))
            
            if file_path:
                page.save(f"{file_path}/Component_plots_collage_page{i+1}.jpg")
            else:
                page.save(f"Plots/Component_plots/Collage/pages/Component_plots_collage_page{i+1}.jpg")

        print("Finished.")
        print("The folder 'Plots/Component_plots/Collage/temp' can be deleted.")
    
    def bmu_template(self, 
                     figsize = (10,10),
                     title_size = 24,
                     fontsize = 10,
                     save = False,
                     file_name = None,
                     file_path = False):
        """
        Generates the BMU map for the current Kohonen map.        

        Args:
                figsize: size of the variable's plot screen.
                    Default: (10,10)

                title_size: size of the plotted title. Default: 24
                    
                fontsize: Default: 10

                save: boolean to define the saving of the created image.
                    This saving will be done in the directory (Plots/Bmu_template).
                    Default: False.

                file_name: the name that will be given to the saved file. If no name
                    is given, the name of the project will be used.

                file_path: the path on the system where the image should be
                    saved in case you do not choose to use the default path.
        """

        # Create coordinates
        xx = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,0], (self.mapsize[1], self.mapsize[0]))
        yy = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,1], (self.mapsize[1], self.mapsize[0]))

        # Plotting
        f = plt.figure(figsize=figsize, dpi=300)
        gs = gridspec.GridSpec(100, 20)
        ax = f.add_subplot(gs[:, 0:20])
        ax.set_aspect('equal')

        # Normalize colors for all hexagons
        norm = mpl.colors.Normalize(vmin=0, vmax=1)
        
        # Create numbering
        nnodes = self.mapsize[0] * self.mapsize[1]
        grid_bmus = np.linspace(1,nnodes, nnodes).reshape(self.mapsize[1], self.mapsize[0])

        # Fill the plot with the hexagons
        for j in range(self.mapsize[1]):
            for i in range(self.mapsize[0]):
                hex = RegularPolygon((xx[(j,i)], yy[(j,i)]), 
                                     numVertices=6, 
                                     radius=1/np.sqrt(3), facecolor=[cm.Pastel1(norm(0.2)) if i%2==0 else cm.Pastel1(norm(0.6))][0],
                                     alpha= [0.3 if j%2==0 else 0.8][0], 
                                     edgecolor='gray'
                                    )
                ax.add_patch(hex)
                
                ax.text(xx[(j,i)], yy[(j,i)], 
                        s=f"{int(grid_bmus[j,i])}", 
                        size = fontsize,
                        horizontalalignment='center', 
                        verticalalignment='center', 
                        color='black')
        
        plt.title("Neurons Template", size=title_size)
        
        # Limites de Plotagem e eixos
        plt.xlim(-0.5, self.mapsize[0]+0.5)
        plt.ylim(-0.5660254, self.mapsize[1]*0.8660254+2*0.560254)
        ax.set_axis_off()
        plt.gca().invert_yaxis()
        
        if save:
            if file_path:
                f.savefig(f"{file_path}/{file_name}_neurons_template.jpg",dpi=300, bbox_inches = "tight")
            else:
                path = 'Plots/Neurons_template'
                os.makedirs(path, exist_ok=True)

                f.savefig(f"Plots/Neurons_template/{file_name}_neurons_template.jpg",dpi=300, bbox_inches = "tight")

    def generate_rec_lattice(self, n_columns, n_rows):
        """
        Generates the xy coordinates of the BMUs for a rectangular grid.

        Args:
            n_rows: Number of lines in the kohonen map.
            n_columns: Number of columns in the kohonen map.

        Returns:
            Coordinates of the [x,y] format for the BMUs in a rectangular grid.

        """
        x_coord = []
        y_coord = []
        for j in range(n_rows):
            for i in range(n_columns):
                x_coord.append(i)
                y_coord.append(j)
        coordinates = np.column_stack([x_coord, y_coord])
        return coordinates

    @property
    def hits_dictionary(self):
        """
        Function to create a hits dictionary from the input vectors for
        each of its BMUs, proportional to the size of the plot.
        """
        # Hits count
        unique, counts = np.unique(self.bmus, return_counts=True)

        # Normalize this count from 0.5 to 2.0 (from a small hexagon to a
        # hexagon that covers half of the neighbors).
        counts = minmax_scale(counts, feature_range = (0.5,2))

        return dict(zip(unique, counts))


    def generate_hex_lattice(self, n_columns, n_rows):
        """
        Generates the xy coordinates of the BMUs for an odd-r hex grid.
        Args:
            n_rows: Number of lines in the kohonen map.
            n_columns: Number of columns in the kohonen map.

        Returns:
            Coordinates in the [x,y] format for the BMUs in a hexagonal grid.

        """
        ratio = np.sqrt(3) / 2

        coord_x, coord_y = np.meshgrid(np.arange(n_columns),
                                       np.arange(n_rows), 
                                       sparse=False, 
                                       indexing='xy')
        coord_y = coord_y * ratio
        coord_x = coord_x.astype(float)
        coord_x[1::2, :] += 0.5
        coord_x = coord_x.ravel()
        coord_y = coord_y.ravel()

        coordinates = np.column_stack([coord_x, coord_y])

        return coordinates



    def plot_torus(self, inner_out_prop = 0.4, red_factor = 4, hits=False):
        """
        Returns the (x, y, z) corrdinates for drawing a toroid.
        
        Args:
        - rows (int): number of grid rows.
        - cols (int): number of grid columns.
        - aspect_ratio (float): proportion between the width and height of the image.
        - R_scale (float): external radius scale of toroid. Default: 0.4.

        Returns:
        (x, y, z) coordinates for drawing a toroid.
        """

        mat_im = self.plot_umatrix(figsize = (10,10), 
                                       hits = True if hits else False, 
                                       save = True, 
                                       file_name = None,
                                       file_path = False, 
                                       resume=True)
            
        # Toroid Resolution
        y_res = int(mat_im.shape[0]/red_factor)
        x_res = int(mat_im.shape[1]/red_factor)

        if mat_im.shape[0] > mat_im.shape[1]:
            mat_im = rotate(mat_im, 90)

        # Reduce image
        mat_res = (resize(mat_im, (y_res, x_res))*256).astype(int)

        def torus(rows, cols, aspect_ratio, R_scale=0.4):
            r_scale = R_scale * aspect_ratio * inner_out_prop
            u, v = np.meshgrid(np.linspace(0, 2*pi, cols), np.linspace(0, 2*pi, rows))
            return (R_scale+r_scale*np.sin(v))*np.cos(u), (R_scale+r_scale*np.sin(v))*np.sin(u), r_scale*np.cos(v)


        r, c, _ = mat_res.shape
        aspect_ratio = mat_im.shape[1]/mat_im.shape[0]
        x, y, z = torus(r, c, aspect_ratio, R_scale=0.4)
        I, J, K, tri_color_intensity, pl_colorscale = self.mesh_data(mat_res, n_colors=32, n_training_pixels=10000) 
        fig5 = go.Figure()
        fig5.add_mesh3d(x=x.flatten(), y=y.flatten(), z=np.flipud(z).flatten(),  
                                    i=I, j=J, k=K, intensity=tri_color_intensity, intensitymode="cell", 
                                    colorscale=pl_colorscale, showscale=False)
        
        scene_style = dict(scene=dict(xaxis_visible=False, yaxis_visible=False, zaxis_visible=False),
                   scene_aspectmode="data")
        
        fig5.update_layout(width=700, height=700,
                        margin=dict(t=10, r=10, b=10, l=10),
                        scene_camera_eye=dict(x=-1.75, y=-1.75, z=1), 
                        **scene_style)
        fig5.show()


    def mesh_data(self, img, n_colors=32, n_training_pixels=800):
        rows, cols, _ = img.shape
        z_data, pl_colorscale = self.image2zvals(img, n_colors=n_colors, n_training_pixels=n_training_pixels)
        triangles = self.regular_tri(rows, cols) 
        I, J, K = triangles.T
        zc = z_data.flatten()[triangles] 
        tri_color_intensity = [zc[k][2] if k%2 else zc[k][1] for k in range(len(zc))]  
        return I, J, K, tri_color_intensity, pl_colorscale
    
    def regular_tri(self, rows, cols):
        #define triangles for a np.meshgrid(np.linspace(a, b, cols), np.linspace(c,d, rows))
        triangles = []
        for i in range(rows-1):
            for j in range(cols-1):
                k = j+i*cols
                triangles.extend([[k,  k+cols, k+1+cols], [k, k+1+cols, k+1]])
        return np.array(triangles) 
    
    def image2zvals(self, img,  n_colors=64, n_training_pixels=800, rngs = 123): 
        # Image color quantization
        # img - np.ndarray of shape (m, n, 3) or (m, n, 4)
        # n_colors: int,  number of colors for color quantization
        # n_training_pixels: int, the number of image pixels to fit a KMeans instance to them
        # returns the array of z_values for the heatmap representation, and a plotly colorscale
    
        if img.ndim != 3:
            raise ValueError(f"Your image does not appear to  be a color image. It's shape is  {img.shape}")
        rows, cols, d = img.shape
        if d < 3:
            raise ValueError(f"A color image should have the shape (m, n, d), d=3 or 4. Your  d = {d}") 
            
        range0 = img[:, :, 0].max() - img[:, :, 0].min()
        if range0 > 1: #normalize the img values
            img = np.clip(img.astype(float)/255, 0, 1)
            
        observations = img[:, :, :3].reshape(rows*cols, 3)
        training_pixels = shuffle(observations, random_state=rngs)[:n_training_pixels]
        model = KMeans(n_clusters=n_colors, random_state=rngs).fit(training_pixels)
        
        codebook = model.cluster_centers_
        indices = model.predict(observations)
        z_vals = indices.astype(float) / (n_colors-1) #normalization (i.e. map indices to  [0,1])
        z_vals = z_vals.reshape(rows, cols)
        # define the Plotly colorscale with n_colors entries    
        scale = np.linspace(0, 1, n_colors)
        colors = (codebook*255).astype(np.uint8)
        pl_colorscale = [[sv, f'rgb{tuple(color)}'] for sv, color in zip(scale, colors)]
        
        # Reshape z_vals  to  img.shape[:2]
        return z_vals.reshape(rows, cols), pl_colorscale



from sklearn.metrics import multilabel_confusion_matrix
from sklearn.metrics import roc_curve, roc_auc_score
import pandas as pd
import numpy as np
import os
from matplotlib import pyplot as plt


class Evaluation(object):
    
    def __init__(self, som_object):
        self.name = som_object.name
        self.pred_size = som_object.pred_size
        self.som = som_object
    
    def evaluation_report(self, 
                          data_test, 
                          labels = None, 
                          bayesian_thresh=False,
                          best_lim = False,
                          save=True, 
                          plot_roc = False,
                          save_roc=False):
        """
        Returns the semi-supervised learning evaluation dataframe.
        Args:
            data_test: Test data, in dataframe or numpy ndarray format.            

            bayesian_thresh: Need to implement. Allows you to perform Bayesian evaluation of
                thresholds.
        """
        
        # Check formats for adaptation
        if isinstance(data_test, pd.DataFrame):
            sample_names = data_test.index.values
            labels = [label for label in data_test.iloc[:,-self.pred_size:].columns]
            data_test = data_test.values
        elif isinstance(data_test, np.ndarray):
            data_test = data_test
            sample_names = [f"Test_sample_{i}" for i in range(1,data_test.shape[0]+1)]
            labels = [f"Var{i}" for i in range(self.pred_size)]
        else:
            print("Only DataFrame and ndarray formats are accepted as input")
        
        # True labels
        y_true = data_test[:,-self.pred_size:]
        # Project values
        y_pred = self.som.project_nan_data(data_test, 
                                           with_labels=True, 
                                           save=False, 
                                           sample_names = sample_names).iloc[:,-self.pred_size:].values
        
        
        
        if best_lim:
            thresh = self.evaluate_thresh(data_test=data_test, 
                                          labels=labels, 
                                          plot=False, 
                                          save=False)
        else:    
            thresh = 0.5 
        
        report_df = self.evaluate(y_pred, y_true, thresh, labels)
        if best_lim:
            report_df["Thresh"] = thresh
        else:
            report_df["Thresh"] = np.array([thresh]*y_true.shape[1])
        
        if save:
            print("Saving...")
            path = 'Evaluation'
            os.makedirs(path, exist_ok=True)
            if best_lim:
                report_df.to_excel(f"Evaluation/Evaluation_SOM_best_lim{self.name}.xlsx")
            else:
                report_df.to_excel(f"Evaluation/Evaluation_SOM_{self.name}.xlsx")
        if plot_roc:
            for i, label in enumerate(labels):
                if report_df["Total Positives"].values[i]!=0:
                    fper, tper, _ = roc_curve(y_true[:,i], y_pred[:,i])
                    roc_score = roc_auc_score(y_true[:,i], y_pred[:,i])
                    plt.figure(figsize=(7,7))
                    plt.plot(fper, tper, color='red', label='ROC')
                    plt.plot([0, 1], [0, 1], color='green', linestyle='--')
                    plt.xlabel('False Positive Rate')
                    plt.ylabel('True Positive Rate')
                    plt.title(f"ROC {label[:10]}")
                    plt.text(0.4, 0.2, f"AUC:{round(roc_score,2)}", fontsize=12)
                    plt.legend()
                    if save_roc:
                        path = 'Evaluation/ROC'
                        os.makedirs(path, exist_ok=True)
                        plt.savefig(f"Evaluation/ROC/ROC_{label[:7]}.png", dpi=200)
                    plt.show()
        
        return report_df
    
    def evaluate(self, y_pred, y_true, thresh, sample_names):
        """
        Function to evaluate a semi-supervised training. The evaluation metrics are:
        [ADD A SHORT DEFINITION OF EACH OF THEM]
        * True Negatives (TN)
        * False Negatives (FN)
        * True Positives (TP)
        * False Positives (FP)
        * Accuracy
        * Total Positives
        * Sensitivity
        * Specificity
        * Accuracy
        * False Positive Rate
        * False Negative Rate

        Args:
            y_pred = label values ​​predicted by the SOM model.

            y_true = True label values ​​from the test set.

            thresh = threshold for identifying labels as positive or negative.
                Accepts number format or a list, if you want to apply a
                list of different thresholds for each classifier.


            sample_names = classifiers names, in the same order as in training
                and testing sets.

        Returns:
            A dataframe with the classifiers in the indexes and the training metrics
            in the columns.
        """

        def predict_label(y_pred, thresh):
            """
            Function to predict training labels from a threshold.
            """
            return np.where(y_pred > thresh, 1, 0)
        def divide_nonzero(num, den):
            """
            Function to divide values ​​by 0. The division value is 0.
            """
            return np.divide(num, 
                             den, 
                             out=np.zeros(num.shape, dtype=float), 
                             where=den!=0)

        if len([thresh]) == 1:
            # Predict the labels according to the threshold
            y_pred_thresh = predict_label(y_pred, thresh)
        elif len([thresh]) > 1:
            y_pred_thresh = np.zeros(y_pred.shape)
            for i, ts in enumerate(thresh):
                y_pred_thresh[i] = predict_label(y_pred[i], ts)

        # Create confusion matrix for each label
        cm = multilabel_confusion_matrix(y_true, y_pred_thresh)

        # Separar TN, FN, TP e FP
        TN = np.array([aval[0][0] for aval in cm])
        FN = np.array([aval[1][0] for aval in cm])
        TP = np.array([aval[1][1] for aval in cm])
        FP = np.array([aval[0][1] for aval in cm])

        # Positives on the true values
        PTV = y_true.sum(axis=0).astype(int)
        # Sensitivity, hit rate, recall, or true positive rate
        TPR = np.around(divide_nonzero(TP, (TP+FN))*100,2)
        # Specificity or true negative rate
        TNR = np.around(divide_nonzero(TN, (TN+FP))*100,2)
        # Precision or positive predictive value
        PPV = np.around(divide_nonzero(TP, (TP+FP))*100,2)
        # Fall out or false positive rate
        FPR = np.around(divide_nonzero(FP, (FP+TN))*100,2)
        # False negative rate
        FNR = np.around(divide_nonzero(FN, (TP+FN))*100,2)
        # Overall accuracy
        ACC = np.around((TP+TN)/(TP+FP+FN+TN)*100,2)
        # Recall
        REC = np.around(divide_nonzero(TP, PTV)*100,2)
        # Create DataFrame
        evaluation = pd.DataFrame(index=sample_names)

        # Preencher DataFrame
        evaluation["True Negatives"] = TN
        evaluation["False Negatives"] = FN
        evaluation["True Positives"] = TP
        evaluation["False Positives"] = FP
        evaluation["Total Positives"] = PTV
        evaluation["Sensitivity"] = TPR
        evaluation["Specificity"] = TNR
        evaluation["Accuracy"] = PPV
        evaluation["Recall"] = REC
        evaluation["False Positive Rate"] = FPR
        evaluation["False Negative Rate"] = FNR
        evaluation["Accuracy"] = ACC


        return evaluation
    
    def evaluate_thresh(self, data_test, labels=None, plot=False, save=False):
        """
        Function to evaluate variation in accuracy, rate of false negatives and rate of
        false positives along the range of thresholds.
        """
        def predict_label(y_pred, thresh):
            """
            Function to predict training labels from a threshold.
            """
            return np.where(y_pred > thresh, 1, 0)
        def divide_nonzero(num, den):
            """
            Function to divide values ​​by 0. The division value is 0.
            """
            return np.divide(num, 
                             den, 
                             out=np.zeros(num.shape, dtype=float), 
                             where=den!=0)
        
        # Check formats for adaptation
        if isinstance(data_test, pd.DataFrame):
            sample_names = data_test.index
            data_test = data_test.values
        elif isinstance(data_test, np.ndarray):
            data_test = data_test
            sample_names = [f"Test_sample_{i}" for i in range(1,data_test.shape[0]+1)]
        else:
            print("Only DataFrame and ndarray formats are accepted as input")
        
        # True labels
        y_true = data_test[:,-self.pred_size:]
        # Project values
        y_pred = self.som.project_nan_data(data_test, with_labels=True, save=False).iloc[:,-self.pred_size:].values
        

        best_lim = np.zeros(self.pred_size)

        for i in range(self.pred_size):
            ACC = np.zeros(len(np.arange(0,1.0001,0.01)))
            FNR = np.zeros(len(np.arange(0,1.0001,0.01)))
            FPR = np.zeros(len(np.arange(0,1.0001,0.01)))
            
            label_index = i
            
            for j,ts in enumerate(np.arange(0,1.0001,0.01)):
                y_pred_thresh = predict_label(y_pred, ts)

                # Criar matriz de confusão para cada label
                cm = multilabel_confusion_matrix(y_true, y_pred_thresh)

                # Separar TN, FN, TP e FP
                TN = np.array([aval[0][0] for aval in cm])
                FN = np.array([aval[1][0] for aval in cm])
                TP = np.array([aval[1][1] for aval in cm])
                FP = np.array([aval[0][1] for aval in cm])

                # Fall out or false positive rate
                FPR[j] = np.around(divide_nonzero(FP, (FP+TN))*100,2)[label_index]
                # False negative rate
                FNR[j] = np.around(divide_nonzero(FN, (TP+FN))*100,2)[label_index]
                # Overall accuracy
                ACC[j] = np.around((TP+TN)/(TP+FP+FN+TN)*100,2)[label_index]
                
            dist = np.zeros(ACC.shape[0])
            for k, (axx, fnr, fpr) in enumerate(zip(ACC, FNR, FPR)):
                dist[k] = abs(fnr-fpr)
                
            best_lim[i] = 0.01 if np.arange(0,1.001,0.01)[np.argmin(dist)] == 0 else np.arange(0,1.001,0.01)[np.argmin(dist)]
            
            if plot:
                #Plot
                plt.figure(figsize=(8,3))
                plt.plot( np.arange(0,1.001,0.01), ACC, label="Accuracy")
                plt.plot(np.arange(0,1.001,0.01), FNR, label="False Negative Rate")
                plt.plot(np.arange(0,1.001,0.01), FPR, label="False Positive Rate")
                plt.vlines(best_lim[i], 0,100, linestyles ="solid", colors ="k")
                plt.title(f"Threshold Evaluation: {labels[label_index]}")
                plt.xlabel("Thresholds")
                plt.ylabel("Rates")
                #plt.xlim(0,0.11)
                plt.legend()
                if save:
                    path = 'Evaluation/best_lim'
                    os.makedirs(path, exist_ok=True)
                    plt.savefig(f"Evaluation/best_lim/best_lim_{labels[label_index][:7]}.png", dpi=200)
            
        return best_lim
            


from sklearn.cluster import KMeans
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
from numpy import nanmean, dot
import matplotlib as mpl
from matplotlib.patches import RegularPolygon
from matplotlib import cm
import os
from math import sqrt
from sklearn.preprocessing import minmax_scale
import numpy as np
from shapely.geometry import Polygon, Point
from shapely.ops import unary_union
import geopandas as gpd
import pkg_resources
from PIL import Image
from tqdm.notebook import tqdm
from sklearn.metrics import davies_bouldin_score as db
import pandas as pd
import plotly.graph_objects as go
from collections import Counter


class ClusterFactory(object):

    def __init__(self, som_object):
        self.som_object = som_object
        self.name = som_object.name
        self.codebook = som_object.codebook.matrix
        self.mapsize = som_object.mapsize
        self.bmus = som_object._bmu[0].astype(int)
        self.neuron_matrix = som_object.neuron_matrix
        self.component_names = som_object.component_names
        self.unit_names = som_object._unit_names
        self.neurons_dataframe = som_object.neurons_dataframe
        self.sample_names = som_object._sample_names
        self.build_umatrix = som_object.build_umatrix
        # Load foot image
        image_file = pkg_resources.resource_filename('intrasom', 'images/foot.jpg')
        self.foot = Image.open(image_file)

    def kmeans(self, k=3, init = "random", n_init=5, max_iter=200):
        """
        Runs the K-means algorithm for grouping data from the trained kohonen map.

        Args:
            k (int, optional): The number of desired clusters. The default is 3.
            init (str, optional): Centroid initialization method. Can be 'random' for random startup
                                or 'k-means++' for smart initialization. The default is 'random'.
            n_init (int, optional): Number of times the K-means algorithm will be executed with different initial
                                    centroids. The final result will be the best obtained among the executions.
                                    The default is 5.
            max_iter (int, optional): Maximum number of iterations of the K-means algorithm for each execution.
                                    The default is 200.

        Returns:
            numpy.ndarray: A two-dimensional array containing the cluster labels assigned to each data point.
                        The labels are in the range [1, k]. The form of the array is (self.mapsize[1], self.mapsize[0]).
        """


        kmeans = KMeans(n_clusters=k, init=init, n_init=n_init, max_iter=max_iter).fit(self.codebook).labels_+1

        return kmeans.reshape(self.mapsize[1], self.mapsize[0])
    
    def Davies_Bouldin_analysis(self, 
                                max_clust = 30, 
                                n_iter=100, 
                                min_type="ensamble", 
                                plot=True, 
                                save=False,
                                verbose=True
                                ):
        """
        Docstring
        """
        def clust_counter(clusts):
            count = np.array(Counter(clusts).most_common())
            count[:,1] = count[:,1]/count[:,1].sum()*100
            return count
        
        # Recovery neuron matrix for analysis
        X = self.neuron_matrix

        # Normalize neuron matrix
        mean = np.mean(X, axis=0)
        std = np.std(X, axis=0)
        X = (X-mean)/std

        # Davies Bouldin Analysis
        n_clusters = np.arange(2, max_clust+1,1)
        db_summary = {}

        for it in tqdm(range(n_iter)):
            db_results = np.zeros(len(n_clusters))
            db_summary[f"Iter{it+1}"] = db_results
            for n_clust in n_clusters:
                kmeans = KMeans(n_clusters=n_clust, 
                                init="random",
                                n_init = 5,
                                max_iter=200
                                ).fit(X)
                db_results[n_clust-2] = db(X, kmeans.labels_)
        
        df = pd.DataFrame.from_dict(db_summary, orient='index', columns=n_clusters)
        if save:
            path = 'Results'
            os.makedirs(path, exist_ok=True)
            df.to_excel(f"Results/{self.name}_dbindex_iterations.xlsx")
        
        # Overwrite X to save space
        X = df.values

        # Minima calculation
        f_mins = []
        s_mins = []
        t_mins = []
        m_mins = []
        max_mins = []
        for it in X:
            x = np.r_[True, it[1:] < it[:-1]] & np.r_[it[:-1] < it[1:], True]
            f_min, s_min, t_min, m_min, max_min = it[x][np.argsort(it[x])[0]], it[x][np.argsort(it[x])[1]],it[x][np.argsort(it[x])[2]], it[x][np.argsort(it[x])[:][int(((len(it[x])-1)/2))]], it[x][np.argsort(it[x])[:][-1]]
            f_mins.append(f_min)
            s_mins.append(s_min)
            t_mins.append(t_min)
            m_mins.append(m_min)
            max_mins.append(max_min)

        f_clust_min = []
        s_clust_min = []
        t_clust_min = []
        m_clust_min = []
        max_clust_min = []
        for i in range(len(f_mins)):
            f_clust_min.append(df.iloc[i][df.iloc[i]==f_mins[i]].index[0])
            s_clust_min.append(df.iloc[i][df.iloc[i]==s_mins[i]].index[0])
            t_clust_min.append(df.iloc[i][df.iloc[i]==t_mins[i]].index[0])
            m_clust_min.append(df.iloc[i][df.iloc[i]==m_mins[i]].index[0])
            max_clust_min.append(df.iloc[i][df.iloc[i]==max_mins[i]].index[0])
        sum_clust_min = f_clust_min + s_clust_min

        if plot:
            x = df.columns.values.astype(int)
            y = df.mean().values
            y_upper = df.quantile(0.75).values
            y_lower = df.quantile(0.25).values
            error = (y_upper-y_lower)/2

            fig = go.Figure()
            for i in range(df.iloc[:,:19].shape[0]):
                fig.add_trace(go.Scatter(
                    x=df.columns.values,
                    y=df.iloc[i],
                    name=f"Iter {i+1}",
                    showlegend=False,
                    line=dict(color="silver")
                ))

            # Confidence Interval
            fig.add_trace(go.Scatter(
                x=x,
                y=y,
                line=dict(color='rgb(0,0,0)'),
                mode='lines',
                name="Trendline", 
                error_y=dict(
                        type='data', # value of error bar given in data coordinates
                        array=error,
                        visible=True)
                ))
                
            fig.update_layout(
                template= "simple_white",
                title= "DBI Comparison: First and Second Global Minima",
                xaxis_title="Cluster Number",
                yaxis_title="Davies-Bouldin Index",
                width=1500,
                height=600,
                font=dict(
                    size=20
                ),
                legend=dict(
                    orientation="h", y=-0.2)
            )

            fig.add_trace(go.Scatter(x=f_clust_min, y=f_mins, mode='markers', name='Global Minimum',
                                        marker_symbol ='circle-open', marker=dict(size=10, color = '#0000FF', line=dict(width=3))))
            fig.add_trace(go.Scatter(x=s_clust_min, y=s_mins, mode='markers', name='Second Global Minimum',
                                        marker_symbol ='circle-open', marker=dict(size=10, color = '#DC3912', line=dict(width=3))))
            #fig.add_trace(go.Scatter(x=t_clust_min, y=t_mins, mode='markers', name='Third to Global Minimum',
            #                            marker_symbol ='circle-open', marker=dict(size=10, color = '#ffff00', line=dict(width=3))))
            #fig.add_trace(go.Scatter(x=m_clust_min, y=m_mins, mode='markers', name='Average Local Minimum',
            #                            marker_symbol ='circle-open', marker=dict(size=10, color = '#ffff00', line=dict(width=3))))
            fig.show()

        # Cluster number counting
        fcounter = clust_counter(f_clust_min)
        scounter = clust_counter(s_clust_min)
        sumcounter = clust_counter(sum_clust_min)

        if verbose:
            print("Davies-Bouldin results:")
            print("")
            print("First Minimum:")
            print(f"Cluster Number: {fcounter[0][0]}. Percentage: {fcounter[0][1]}%.")
            print(f"Cluster Number: {fcounter[1][0]}. Percentage: {fcounter[1][1]}%.")
            print("")
            print("Second Minimum:")
            print(f"Cluster Number: {scounter[0][0]}. Percentage: {scounter[0][1]}%.")
            print(f"Cluster Number: {scounter[1][0]}. Percentage: {scounter[1][1]}%.")
            print("")
            print("Ensamble Minimum:")
            print(f"Cluster Number: {sumcounter[0][0]}. Percentage: {sumcounter[0][1]}%.")
            print(f"Cluster Number: {sumcounter[1][0]}. Percentage: {sumcounter[1][1]}%.")
        else:
            print(f"Davies-Bouldin results:")
            print(f"First Minimum: {fcounter[0][0]}")
            print(f"Second Minimum: {scounter[0][0]}")
            print(f"Ensamble Minimum: {sumcounter[0][0]}")
    
    def results_cluster(self, clusters, save=True, savetype="parquet"):
        """
        Returns a DataFrame containing the results of grouping the data, with the addition of the assigned cluster labels.

        Args:
            clusters (numpy.ndarray): A one-dimensional array containing the cluster labels assigned to each data point.
            save (bool, optional): Indicates whether the results should be saved in an Excel file. The default is True.

        Returns:
            pandas.DataFrame: A DataFrame containing the results of clustering the data, including the assigned cluster labels.
        """

        df = self.neurons_dataframe.copy()
        df[f"{clusters.max()}_clusters"] = clusters.flatten()
        df = df.iloc[self.bmus,:]

        df.set_index(self.sample_names, inplace=True)

        if save:
            path = 'Results'
            os.makedirs(path, exist_ok=True)
            if savetype == "xlsx":
                df.to_excel(f"Results/{self.name}_results_{clusters.max()}_clusters.xlsx")
            elif savetype == "parquet":
                df.to_parquet(f"Results/{self.name}_results_{clusters.max()}_clusters.parquet")

        return df
            
    def plot_kmeans(self, 
                    clusters, 
                    figsize = (16,14),
                    title_size = 25,
                    title_pad = 40,
                    legend_title = False,
                    legend_text_size = 10,
                    save = False,
                    file_name = None,
                    file_path = False,
                    watermark_neurons = False, 
                    neurons_fontsize = 12,
                    umatrix = False, 
                    hits=False, 
                    alfa_clust = 0.5, 
                    log=False,
                    colormap = "gist_rainbow",
                    clusters_highlight = [],
                    legend_title_size=12,
                    cluster_outline=False,
                    plot_labels=False,
                    custom_labels = [],
                    clusterout_maxtext_size=12,
                    return_geodataframe=False,
                    auto_adjust_text =False):
        """
        Plots a graph with the clusters resulting from the execution of the K-means algorithm.

        Parameters:
        - clusters: ndarray or 2-dimensional list
            Array or 2-dimensional list containing the clusters generated by the K-means algorithm.
        - figsize: tuple, optional
            Figure size (width, height). Default is (16, 14).
        - title_size: int, optional
            Font size of the title. Default is 25.
        - title_pad: int, optional
            Spacing between the title and the top of the graph. Default is 40.
        - legend_text_size: int, optional
            Font size of the legend text. Default is 10.
        - save: bool, optional
            Indicates whether the graph should be saved to a file. Default is False.
        - file_name: str, optional
            Name of the file to be saved. Default is None.
        - file_path: bool, optional
            Indicates whether the file path should be included when saving. Default is False.
        - watermark_neurons: bool, optional
            Indicates whether the BMU numbers should be displayed as a watermark on the graph. Default is False.
        - neurons_fontsize: int, optional
            Font size of the BMU numbers. Default is 12.
        - umatrix: bool, optional
            Indicates whether the U-matrix should be plotted on the graph. Default is False.
        - hits: bool, optional
            Indicates whether the hits should be plotted on the graph. Default is False.
        - alfa_clust: float, optional
            Transparency value of the clusters in the graph. Default is 0.5.
        - log: bool, optional
            Indicates whether the U-matrix should be plotted on a logarithmic scale. Default is False.
        - colormap: str, optional
            Name of the colormap to be used in the graph. Default is "gist_rainbow".
        - clusters_highlight: list, optional
            List containing the clusters that should be highlighted in the graph. Default is [].
        - legend_title_size: int, optional
            Font size of the legend title. Default is 12.
        - cluster_outline: bool, optional
            Indicates whether the cluster outlines should be drawn on the graph. Default is False.
        - plot_labels: bool, optional
            Indicates whether the labels should be plotted on the graph. Default is False.
        - custom_labels: list, optional
            List containing custom labels for each point on the graph. Default is [].
        - clusterout_maxtext_size: int, optional
            Maximum size of the cluster text on the graph. Default is 12.
        - return_geodataframe: bool, optional
            Indicates whether the GeoDataFrame should be returned. Default is False.
        - auto_adjust_text: bool, optional
            Indicates whether automatic text adjustment should be enabled. Default is False.
        """
        

        if file_name is None:
            file_name = f"Clusters_{len(np.unique(clusters))}_{self.name}"


        f = plt.figure(figsize=figsize, dpi=300)
        gs = gridspec.GridSpec(100, 100)
        
        max_clust = clusters.max() if len(clusters_highlight) == 0 else len(clusters_highlight)
        max_clust = max_clust+1 if watermark_neurons else max_clust

        pad_subplots = 3
        if max_clust <=10:
            ax = f.add_subplot(gs[:95, :90-pad_subplots])
        elif max_clust<=20:
            ax = f.add_subplot(gs[:95, :80-pad_subplots])
        elif max_clust<=30:
            ax = f.add_subplot(gs[:95, :70-pad_subplots])
        else:
            ax = f.add_subplot(gs[:95, :60-pad_subplots])

        ax.set_aspect('equal')

        xx = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,0], (self.mapsize[1], self.mapsize[0]))
        yy = np.reshape(self.generate_hex_lattice(self.mapsize[0], self.mapsize[1])[:,1], (self.mapsize[1], self.mapsize[0]))

        if umatrix:
            if hits:
                bmu_dic = self.hits_dictionary
            # U Matrix
            um = self.build_umatrix(expanded = True, log=log)
            umat = self.build_umatrix(expanded = False, log=log)

            # Normalize the colors for all hexagons
            norm = mpl.colors.Normalize(vmin=np.nanmin(um), vmax=np.nanmax(um))
            counter = 0

            alfa = 1

            for j in range(self.mapsize[1]):
                for i in range(self.mapsize[0]):
                    # Central Hexagon
                    hexagon = RegularPolygon((xx[(j, i)]*2,
                                          yy[(j,i)]*2),
                                         numVertices=6,
                                         radius=1/np.sqrt(3),
                                         facecolor= cm.jet(norm(umat[j][i])),
                                         alpha=alfa, 
                                         zorder=0)#, edgecolor='black')

                    ax.add_patch(hexagon)

                    # Right Hexagon
                    if not np.isnan(um[j, i, 0]):
                        hexagon = RegularPolygon((xx[(j, i)]*2+1,
                                              yy[(j,i)]*2),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,0])),
                                             alpha=alfa, 
                                             zorder=0)
                        ax.add_patch(hexagon)

                    # Upper Right Hexagon
                    if not np.isnan(um[j, i, 1]):
                        hexagon = RegularPolygon((xx[(j, i)]*2+0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,1])),
                                             alpha=alfa, 
                                             zorder=0)
                        ax.add_patch(hexagon)

                    # Upper Left Hexagon
                    if not np.isnan(um[j, i, 2]):
                        hexagon = RegularPolygon((xx[(j, i)]*2-0.5,
                                              yy[(j,i)]*2+(np.sqrt(3)/2)),
                                             numVertices=6,
                                             radius=1/np.sqrt(3),
                                             facecolor=cm.jet(norm(um[j,i,2])),
                                             alpha=alfa, 
                                             zorder=0)
                        ax.add_patch(hexagon)

                    # Plot hits
                    if hits:
                        try:
                            hexagon = RegularPolygon((xx[(j, i)]*2,
                                                  yy[(j,i)]*2),
                                                 numVertices=6,
                                                 radius=((1/np.sqrt(3))*bmu_dic[counter]),
                                                 facecolor='white',
                                                 alpha=alfa, 
                                                 zorder=0)
                            ax.add_patch(hexagon)
                        except:
                            pass
                    counter+=1
        
        norm = mpl.colors.Normalize(vmin=np.nanmin(clusters), vmax=np.nanmax(clusters))

        cmap = cm.get_cmap(colormap)
        if cluster_outline:
            cluster_vertices_dict = {}

            for j in range(clusters.shape[0]):
                for i in range(clusters.shape[1]):

                    label = clusters[j][i]

                    if len(clusters_highlight) == 0:
                        color = cmap(norm(clusters[j][i]))
                    else:   
                        color = "gray" if clusters[j][i] not in clusters_highlight else cmap(norm(clusters[j][i]))

                    if label not in cluster_vertices_dict:
                        cluster_vertices_dict[label] = []
                    
                    # Get the vertices of the hexagon
                    hexagon = RegularPolygon((xx[(j, i)]*2, yy[(j,i)]*2), 
                                        numVertices=6, 
                                        radius=2/np.sqrt(3)+0.04,
                                        facecolor = color, 
                                        alpha = alfa_clust)
                    vertices = hexagon.get_verts()
                    polygon = Polygon(vertices)
                    cluster_vertices_dict[label].append(polygon)

                    if watermark_neurons:
                        # For use in BMUs number plot
                        nnodes = self.mapsize[0] * self.mapsize[1]
                        grid_bmus = np.linspace(1,nnodes, nnodes).reshape(self.mapsize[1], self.mapsize[0])
                        # Central Hexagon
                        hexagon = RegularPolygon((xx[(j, i)]*2,
                                                yy[(j,i)]*2),
                                                numVertices=6,
                                                radius=2/np.sqrt(3),
                                                facecolor= "white",
                                                edgecolor='black',
                                                alpha=0.1, 
                                                zorder=2)#, edgecolor='black')
                        ax.add_patch(hexagon)

                        ax.text(xx[(j,i)]*2, yy[(j,i)]*2, 
                                s=f"{int(grid_bmus[j,i])}", 
                                size = neurons_fontsize,
                                horizontalalignment='center', 
                                verticalalignment='center', 
                                color='black', 
                                zorder=2)
        else:
            for j in range(clusters.shape[0]):
                for i in range(clusters.shape[1]):

                    if len(clusters_highlight) == 0:
                        color = cmap(norm(clusters[j][i]))
                    else:
                        color = "gray" if clusters[j][i] not in clusters_highlight else cmap(norm(clusters[j][i]))

                    hexagon = RegularPolygon((xx[(j, i)]*2, yy[(j,i)]*2), 
                                        numVertices=6, 
                                        radius=2/np.sqrt(3)-0.04*(2/np.sqrt(3)),
                                        facecolor = color, 
                                        alpha = alfa_clust, 
                                         zorder=1)
                    ax.add_patch(hexagon)                
                    
                    hexagon = RegularPolygon((xx[(j, i)]*2, yy[(j,i)]*2), 
                                        numVertices=6, 
                                        radius=2/np.sqrt(3)-0.04*(2/np.sqrt(3)),
                                        fill=False,
                                        facecolor = None, 
                                        edgecolor = color,
                                        linewidth=1.9, 
                                        alpha = 1, 
                                        zorder = 1)
                    ax.add_patch(hexagon)
                
                    if watermark_neurons:
                        # For use in neurons number plot
                        nnodes = self.mapsize[0] * self.mapsize[1]
                        grid_bmus = np.linspace(1,nnodes, nnodes).reshape(self.mapsize[1], self.mapsize[0])
                        # Central Hexagon
                        hexagon = RegularPolygon((xx[(j, i)]*2,
                                                yy[(j,i)]*2),
                                                numVertices=6,
                                                radius=2/np.sqrt(3),
                                                facecolor= "white",
                                                edgecolor='black',
                                                alpha=0.1, 
                                                zorder=2)#, edgecolor='black')
                        ax.add_patch(hexagon)

                        ax.text(xx[(j,i)]*2, yy[(j,i)]*2, 
                                s=f"{int(grid_bmus[j,i])}", 
                                size = neurons_fontsize,
                                horizontalalignment='center', 
                                verticalalignment='center', 
                                color='black', 
                                zorder=2)

        if cluster_outline:
            cluster_vertices_dict = dict(sorted(cluster_vertices_dict.items()))

            # Create a list of GeoSeries for each cluster
            cluster_geo_series = []
            for label in cluster_vertices_dict:
                cluster_geo_series.append(gpd.GeoSeries(cluster_vertices_dict[label]))

            # Dissolve the polygons in each GeoSeries
            dissolved_geometries = []
            for geo_series in cluster_geo_series:
                dissolved_geometry = unary_union(geo_series)
                dissolved_geometries.append(dissolved_geometry)

            # Colors of the clusters
            colors = []
            labels_default = []
            cluster = []
            for i in range(len(dissolved_geometries)):
                label = i+1
                if len(clusters_highlight) == 0:
                    color = cmap(norm(label))
                else:
                    color = "gray" if label not in clusters_highlight else cmap(norm(label))
                
                colors.append(color)
                labels_default.append(f"#{label}")
                cluster.append(label)

            # Create a GeoDataFrame from the list of geometries
            gdf = gpd.GeoDataFrame(geometry=[geom for geom in dissolved_geometries])

            # Buffer inside to plot the outline
            gdf["geometry"] = gdf.buffer(-0.075)

            # Add a new column to the GeoDataFrame with the colors
            gdf['color'] = colors
            gdf['label'] = labels_default
            gdf['cluster'] = cluster

            if len(custom_labels)>0:
                gdf['label'] = custom_labels

            gdf = gdf.explode()

            # Plot the geometry
            gdf.plot(ax=ax, 
                     facecolor=gdf['color'], 
                     edgecolor='none', 
                     alpha=alfa_clust,
                     zorder=1)
            gdf.plot(ax=ax, 
                     facecolor='none', 
                     edgecolor=gdf['color'], 
                     alpha=1, 
                     linewidth=2, 
                     zorder=1)
            if plot_labels:
                # Iterate over each polygon in the GeoDataFrame
                for idx, row in gdf.iterrows():
                    
                    if len(clusters_highlight)!=0:
                        if row["cluster"] not in clusters_highlight:
                            continue
                    
                    if auto_adjust_text:
                    
                        polygon = row['geometry']
                        label = row['label']
                        
                        # calculate the minimum bounding box of the polygon
                        mbb = Polygon(polygon.exterior).minimum_rotated_rectangle

                        # Get the minimum rotated rectangle of the bounding box
                        rotated_rect = mbb.minimum_rotated_rectangle

                        # Get the angle of the major axis of the minimum rotated rectangle
                        angle = np.rad2deg(np.arctan2(rotated_rect.bounds[3]-rotated_rect.bounds[1],
                                                      rotated_rect.bounds[2]-rotated_rect.bounds[0]))

                        x, y = mbb.representative_point().coords[0]
                        ax.plot(*mbb.exterior.xy, color="red")
                        mbb_coords = mbb.exterior.coords
                        
                        # calculate the aspect ratio of the MBB
                        mbb_width = Point(mbb_coords[0]).distance(Point(mbb_coords[1]))
                        mbb_height = Point(mbb_coords[1]).distance(Point(mbb_coords[2]))

                        if mbb_width<mbb_height:
                            save_wid = mbb_width
                            mbb_width = mbb_height
                            mbb_height = save_wid
                        
                        plt.ioff()
                        # calculate the aspect ratio of the label text
                        label_width, label_height = ax.text(0, 0, label, ha='left', va='bottom', fontsize=clusterout_maxtext_size).get_window_extent().size
                        plt.ion()

                        # scale the label text to fit inside the MBB
                        if label_width>mbb_width:
                            scale_factor = mbb_width/label_width
                        else:
                            scale_factor = 1

                        fontsize = clusterout_maxtext_size * scale_factor*10
                        
                        
                        ax.text(x, 
                                y, 
                                label, 
                                ha='center', 
                                va='center', 
                                fontsize=fontsize, 
                                rotation=angle, 
                                color='white', 
                                weight='bold')
                    else:
                        # Get the centroid of the polygon
                        centroid = row.geometry.centroid
                        
                        # Get the label
                        label = row['label']
                        
                        # Create a text object with the label
                        ax.text(x=centroid.x+0.05, 
                                y=centroid.y+0.05, 
                                s=label, 
                                ha='center', 
                                va='center', 
                                color='black', 
                                alpha=0.7, 
                                weight='bold', 
                                fontsize=clusterout_maxtext_size)
                        
                        ax.text(x=centroid.x, 
                                y=centroid.y, 
                                s=label, 
                                ha='center', 
                                va='center', 
                                color="white", 
                                weight='bold', 
                                fontsize=clusterout_maxtext_size)
                        


        # Plotting Parameters
        ax.set_xlim(-0.6-0.5, 2*self.mapsize[0]-0.5+0.6)
        ax.set_ylim(-0.5660254-0.81, 2*self.mapsize[1]*0.8660254-2*0.560254+0.75)
        ax.set_axis_off()
        ax.invert_yaxis()

        plt.title(f"Clustering Matrix - {clusters.max()} clusters",
                  horizontalalignment='center',  
                  verticalalignment='top', 
                  size=title_size, 
                  pad=title_pad)
        
        if max_clust <=10:
            ax2 = f.add_subplot(gs[20:80, 90:])
        elif max_clust<=20:
            ax2 = f.add_subplot(gs[20:80, 80:])
        elif max_clust<=30:
            ax2 = f.add_subplot(gs[20:80, 70:])
        else:
            ax2 = f.add_subplot(gs[20:80, 60:])
        
        
        ax2.invert_yaxis()
        ax2.set_aspect('equal')

        n_cols = int(np.ceil(max_clust/10))
        n_rows = int(np.ceil(max_clust / n_cols))

        hex_height = 0.096
        pad = (0.1-hex_height)
        radius =  hex_height/2
        total_height = hex_height * n_rows + n_rows * pad
        shift = hex_height
        y_start = ((1 - total_height) / 2)+shift/2
        x_start = pad+shift/2
        text_pad = hex_height*3

        condition = max_clust-1 if watermark_neurons else max_clust

        for i, (xfac, yfac) in enumerate(np.ndindex((n_cols, n_rows))):
            if i+1 <= condition:
                cluster = i+1 if len(clusters_highlight) == 0 else clusters_highlight[i]
                x_center = x_start+(xfac)*shift+xfac*pad+xfac*text_pad
                y_center = y_start+(yfac)*shift+yfac*pad

                if len(clusters_highlight) == 0:
                    color = cmap(norm(cluster))
                else:
                    color = "gray" if cluster not in clusters_highlight else cmap(norm(cluster))

                hex_points = RegularPolygon((x_center, y_center), 
                                            numVertices=6, 
                                            radius=radius,
                                            facecolor=color, 
                                            edgecolor=None, 
                                            alpha=alfa_clust)
                ax2.add_patch(hex_points)

                hex_points = RegularPolygon((x_center, y_center), 
                                            numVertices=6, 
                                            radius=radius-radius*0.05,
                                            facecolor=None,
                                            fill=False, 
                                            edgecolor=color,
                                            linewidth=2)
                ax2.add_patch(hex_points)
                if len(custom_labels)>0:
                    cluster_name = custom_labels[cluster-1]
                else:
                    cluster_name = f"Cluster #{cluster}"
                ax2.annotate(cluster_name,
                            xy=(x_center+radius+0.01, y_center),
                            xytext=(0, 0),
                            textcoords="offset points",
                            color='black',
                            weight='bold',
                            fontsize=legend_text_size,
                            ha='left',
                            va='center')
            else:
                if watermark_neurons:
                    x_center = x_start+(xfac)*shift+xfac*pad+xfac*text_pad
                    y_center = y_start+(yfac)*shift+yfac*pad

                    hex_points = RegularPolygon((x_center, y_center), 
                                            numVertices=6, 
                                            radius=radius-radius*0.05,
                                            facecolor="White",
                                            fill = True, 
                                            edgecolor="Black",
                                            linewidth=2)
                    
                    ax2.add_patch(hex_points)

                    ax2.annotate(f"#",
                            xy=(x_center, y_center),
                            xytext=(0, 0),
                            textcoords="offset points",
                            color='black',
                            weight='bold',
                            fontsize=legend_text_size,
                            ha='center',
                            va='center')

                    ax2.annotate(f"Neuron Number",
                            xy=(x_center+radius+0.01, y_center),
                            xytext=(0, 0),
                            textcoords="offset points",
                            color='black',
                            weight='bold',
                            fontsize=legend_text_size,
                            ha='left',
                            va='center')
                    break

        
        ax2.set_title(legend_title if legend_title!=False else "Legend", 
                      fontdict={"fontsize": legend_title_size},
                      loc="center", 
                      pad=5,
                      fontweight='bold',
                      y=1-y_start+0.03)
        
        ax2.set_xlim(0, n_cols*hex_height+n_cols*pad+n_cols*text_pad)
        ax2.set_ylim(1, -0.01)
        
        ax2.set_axis_off()

        #ADD WATERMARK
        # Add white space subplot below the plot
        ax3 = f.add_subplot(gs[95:100, 0:20], zorder=-1)

        # Add the watermark image to the white space subplot
        ax3.imshow(self.foot, aspect='equal', alpha=1)
        ax3.axis('off')

        f.subplots_adjust(wspace=0.1)

        if save:
            if file_path:
                f.savefig(f"{file_path}/{file_name}.jpg",dpi=300, bbox_inches = "tight")
            else:
                path = 'Plots/Clusters'
                os.makedirs(path, exist_ok=True)

                f.savefig(f"Plots/Clusters/{file_name}.jpg",dpi=300, bbox_inches = "tight")
        
        if return_geodataframe:
            return gdf
                

    def generate_hex_lattice(self, n_columns, n_rows):
        """
        Generates the xy coordinates of the BMUs for an odd-r hexagonal grid.
        Args:
            n_rows: Number of lines in the Kohonen map.
            n_columns: Number of columns in the Kohonen map.

        Returns:
            Coordinates of the [x,y] format for the BMUs in a hexagonal grid.

        """
        ratio = np.sqrt(3) / 2

        coord_x, coord_y = np.meshgrid(np.arange(n_columns),
                                       np.arange(n_rows), 
                                       sparse=False, 
                                       indexing='xy')
        coord_y = coord_y * ratio
        coord_x = coord_x.astype(float)
        coord_x[1::2, :] += 0.5
        coord_x = coord_x.ravel()
        coord_y = coord_y.ravel()

        coordinates = np.column_stack([coord_x, coord_y])

        return coordinates
    

    # def build_umatrix(self, expanded=False, log=False):
    #     """
    #     Function to calculate the U Matrix of unified distances from the
    #     trained weight matrix.
    # 
    #     Args:
    #         expanded: boolean value to indicate whether the return will be from the summarized
    #             or unified matrix of distances (average of distances from the 6
    #             neighborhood BMUs) or expanded (all distance values)
    #             
    #     Returns:
    #         Expanded or summarized unified distance matrix.
    #     """
    #     # Function to find distance quickly
    #     def fast_norm(x):
    #         """
    #         Returns the L2 norm of a 1-D array.
    #         """
    #         return sqrt(dot(x, x.T))
    # 
    #     # Matrix of BMUs weights
    #     weights = np.reshape(self.codebook, (self.mapsize[1], self.mapsize[0], self.codebook.shape[1]))
    # 
    #     # Neighbor hexagonal search
    #     ii = [[1, 1, 0, -1, 0, 1], [1, 0,-1, -1, -1, 0]]
    #     jj = [[0, 1, 1, 0, -1, -1], [0, 1, 1, 0, -1, -1]]
    # 
    #     # Initialize U Matrix
    #     um = np.nan * np.zeros((weights.shape[0], weights.shape[1], 6))
    # 
    #     # Fill U Matrix
    #     for y in range(weights.shape[0]):
    #         for x in range(weights.shape[1]):
    #             w_2 = weights[y, x]
    #             e = y % 2 == 0
    #             for k, (i, j) in enumerate(zip(ii[e], jj[e])):
    #                 if (x+i >= 0 and x+i < weights.shape[1] and y+j >= 0 and y+j < weights.shape[0]):
    #                     w_1 = weights[y+j, x+i]
    #                     um[y, x, k] = fast_norm(w_2-w_1)
    #     if expanded:
    #         # Expanded U matrix
    #         return np.log(um) if log else um
    #     else:
    #         # Reduced U matrix
    #         return nanmean(np.log(um), axis=2) if log else nanmean(um, axis=2)
        

    @property
    def hits_dictionary(self):
        """
        Function to create a dictionary of hits from the input vectors for
        each of its BMUs, proportional to the size of the plot.
        """
        # Hit count
        unique, counts = np.unique(self.bmus, return_counts=True)

        # Normalize this count from 0.5 to 2.0 (from a small hexagon to a
        # hexagon that covers half of the neighbors).
        counts = minmax_scale(counts, feature_range = (0.5,2))

        return dict(zip(unique, counts))

# K-Means error
os.environ["OMP_NUM_THREADS"] = "2"
